from dataclasses import dataclass
from typing import Any

import pytest

from dsp_tools.utils.connection import Connection

# ruff: noqa: ARG002 (unused-method-argument)


@dataclass
class ConnectionMockBase(Connection):
    """
    Base mock for the Connection class.
    Does not have any capabilities, but subclasses can override methods.
    """

    def get(self, route: str, headers: dict[str, str] | None = None) -> dict[str, Any]:
        raise AssertionError("GET not implemented in mock")

    def put(
        self,
        route: str,
        data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        raise AssertionError("PUT not implemented in mock")

    def post(
        self,
        route: str,
        data: dict[str, Any] | None = None,
        files: dict[str, tuple[str, Any]] | None = None,
        headers: dict[str, str] | None = None,
        timeout: int | None = None,
    ) -> dict[str, Any]:
        raise AssertionError("POST not implemented in mock")

    def delete(
        self,
        route: str,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        raise AssertionError("DELETE not implemented in mock")

    def get_token(self) -> str:
        raise AssertionError("get_token not implemented in mock")

    def login(self, email: str, password: str) -> None:
        raise AssertionError("login not implemented in mock")

    def logout(self) -> None:
        raise AssertionError("logout not implemented in mock")


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

import warnings
from dataclasses import dataclass
from dataclasses import field
from typing import Any

from lxml import etree

from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.models.exceptions import InputError
from dsp_tools.xmllib.models.migration_metadata import MigrationMetadata
from dsp_tools.xmllib.models.values import ColorValue
from dsp_tools.xmllib.models.values import LinkValue
from dsp_tools.xmllib.models.values import Richtext
from dsp_tools.xmllib.value_checkers import find_geometry_problem
from dsp_tools.xmllib.value_checkers import is_color
from dsp_tools.xmllib.value_checkers import is_decimal
from dsp_tools.xmllib.value_checkers import is_string_like

XML_NAMESPACE_MAP = {None: "https://dasch.swiss/schema", "xsi": "http://www.w3.org/2001/XMLSchema-instance"}
DASCH_SCHEMA = "{https://dasch.swiss/schema}"

LIST_SEPARATOR = "\n    - "


@dataclass
class AnnotationResource:
    res_id: str
    label: str
    annotation_of: str
    comments: list[str]
    permissions: str = "res-default"
    migration_metadata: MigrationMetadata | None = None

    def __post_init__(self) -> None:
        _check_strings(string_to_check=self.res_id, res_id=self.res_id, field_name="Resource ID")
        _check_strings(string_to_check=self.label, res_id=self.res_id, field_name="Label")

    def new(
        self, res_id: str, label: str, annotation_of: str, comments: list[str], permissions: str = "res-default"
    ) -> AnnotationResource:
        return AnnotationResource(
            res_id=res_id,
            label=label,
            annotation_of=annotation_of,
            comments=comments,
            permissions=permissions,
        )

    def add_comment(self, comment: str) -> AnnotationResource:
        self.comments.append(comment)
        return self

    def add_comments(self, comments: list[str]) -> AnnotationResource:
        self.comments.extend(comments)
        return self

    def add_migration_metadata(
        self, creation_date: str | None, iri: str | None = None, ark: str | None = None
    ) -> AnnotationResource:
        if self.migration_metadata:
            raise InputError(
                f"The resource with the ID '{self.res_id}' already contains migration metadata, "
                f"no new data can be added."
            )
        self.migration_metadata = MigrationMetadata(creation_date=creation_date, iri=iri, ark=ark, res_id=self.res_id)
        return self

    def serialise(self) -> etree._Element:
        self.comments = _transform_unexpected_input(self.comments, "comments", self.res_id)
        res_ele = self._serialise_resource_element()
        res_ele.append(self._serialise_annotation_of())
        res_ele.append(_serialise_has_comment(self.comments, self.res_id))
        return res_ele

    def _serialise_resource_element(self) -> etree._Element:
        attribs = {"label": self.label, "id": self.res_id}
        if self.permissions:
            attribs["permissions"] = self.permissions
        return etree.Element(f"{DASCH_SCHEMA}annotation", attrib=attribs, nsmap=XML_NAMESPACE_MAP)

    def _serialise_annotation_of(self) -> etree._Element:
        return LinkValue(value=self.annotation_of, prop_name="isAnnotationOf", resource_id=self.res_id).serialise()


@dataclass
class RegionResource:
    res_id: str
    label: str
    color: str
    region_of: str
    geometry: dict[str, Any]
    comments: list[str]
    permissions: str = "res-default"
    migration_metadata: MigrationMetadata | None = None

    def __post_init__(self) -> None:
        _check_strings(string_to_check=self.res_id, res_id=self.res_id, field_name="Resource ID")
        _check_strings(string_to_check=self.label, res_id=self.res_id, field_name="Label")
        if not is_color(self.color):
            msg = (
                f"The color value '{self.color}' for the resource with the ID: '{self.res_id}' failed validation. "
                f"Please consult the documentation for details."
            )
            warnings.warn(DspToolsUserWarning(msg))
        if fail_msg := find_geometry_problem(self.geometry):
            msg = f"The geometry of the resource with the ID '{self.res_id}' failed validation.\n" + fail_msg
            warnings.warn(DspToolsUserWarning(msg))

    def new(
        self,
        res_id: str,
        label: str,
        color: str,
        region_of: str,
        geometry: dict[str, Any],
        comments: list[str],
        permissions: str = "res-default",
    ) -> RegionResource:
        return RegionResource(
            res_id=res_id,
            label=label,
            color=color,
            region_of=region_of,
            geometry=geometry,
            comments=comments,
            permissions=permissions,
        )

    def add_comment(self, comment: str) -> RegionResource:
        self.comments.append(comment)
        return self

    def add_comments(self, comments: list[str]) -> RegionResource:
        self.comments.extend(comments)
        return self

    def add_migration_metadata(
        self, creation_date: str | None, iri: str | None = None, ark: str | None = None
    ) -> RegionResource:
        if self.migration_metadata:
            raise InputError(
                f"The resource with the ID '{self.res_id}' already contains migration metadata, "
                f"no new data can be added."
            )
        self.migration_metadata = MigrationMetadata(creation_date=creation_date, iri=iri, ark=ark, res_id=self.res_id)
        return self

    def serialise(self) -> etree._Element:
        self.comments = _transform_unexpected_input(self.comments, "comments", self.res_id)
        res_ele = self._serialise_resource_element()
        res_ele.append(self._serialise_geometry())
        res_ele.extend(self._serialise_values())
        if self.comments:
            res_ele.append(_serialise_has_comment(self.comments, self.res_id))
        return res_ele

    def _serialise_resource_element(self) -> etree._Element:
        attribs = {"label": self.label, "id": self.res_id}
        if self.permissions:
            attribs["permissions"] = self.permissions
        return etree.Element(f"{DASCH_SCHEMA}region", attrib=attribs, nsmap=XML_NAMESPACE_MAP)

    def _serialise_values(self) -> list[etree._Element]:
        return [
            ColorValue(value=self.color, prop_name="hasColor", resource_id=self.res_id).serialise(),
            LinkValue(value=self.region_of, prop_name="isRegionOf", resource_id=self.res_id).serialise(),
        ]

    def _serialise_geometry(self) -> etree._Element:
        geo_prop = etree.Element(f"{DASCH_SCHEMA}geometry-prop", name="hasGeometry", nsmap=XML_NAMESPACE_MAP)
        ele = etree.Element(f"{DASCH_SCHEMA}geometry", nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.geometry)
        geo_prop.append(ele)
        return geo_prop


@dataclass
class LinkResource:
    res_id: str
    label: str
    link_to: list[str]
    comments: list[str]
    permissions: str = "res-default"
    migration_metadata: MigrationMetadata | None = None

    def new(
        self, res_id: str, label: str, link_to: list[str], comments: list[str], permissions: str = "res-default"
    ) -> LinkResource:
        return LinkResource(
            res_id=res_id,
            label=label,
            link_to=link_to,
            comments=comments,
            permissions=permissions,
        )

    def add_comment(self, comment: str) -> LinkResource:
        self.comments.append(comment)
        return self

    def add_comments(self, comments: list[str]) -> LinkResource:
        self.comments.extend(comments)
        return self

    def add_migration_metadata(
        self, creation_date: str | None, iri: str | None = None, ark: str | None = None
    ) -> LinkResource:
        if self.migration_metadata:
            raise InputError(
                f"The resource with the ID '{self.res_id}' already contains migration metadata, "
                f"no new data can be added."
            )
        self.migration_metadata = MigrationMetadata(creation_date=creation_date, iri=iri, ark=ark, res_id=self.res_id)
        return self

    def serialise(self) -> etree._Element:
        self._check_for_and_convert_unexpected_input()
        res_ele = self._serialise_resource_element()
        res_ele.append(_serialise_has_comment(self.comments, self.res_id))
        res_ele.append(self._serialise_links())
        return res_ele

    def _check_for_and_convert_unexpected_input(self) -> None:
        _check_strings(string_to_check=self.res_id, res_id=self.res_id, field_name="Resource ID")
        _check_strings(string_to_check=self.label, res_id=self.res_id, field_name="Label")
        self.comments = _transform_unexpected_input(self.comments, "comments", self.res_id)
        self.link_to = _transform_unexpected_input(self.link_to, "link_to", self.res_id)

    def _serialise_resource_element(self) -> etree._Element:
        attribs = {"label": self.label, "id": self.res_id}
        if self.permissions:
            attribs["permissions"] = self.permissions
        return etree.Element(f"{DASCH_SCHEMA}link", attrib=attribs, nsmap=XML_NAMESPACE_MAP)

    def _serialise_links(self) -> etree._Element:
        prop_ele = etree.Element(f"{DASCH_SCHEMA}resptr-prop", name="hasLinkTo", nsmap=XML_NAMESPACE_MAP)
        vals = [LinkValue(value=x, prop_name="hasLinkTo", resource_id=self.res_id) for x in self.link_to]
        prop_ele.extend([v.make_element() for v in vals])
        return prop_ele


@dataclass
class SegmentBounds:
    segment_start: float | int | str
    segment_end: float | int | str
    res_id: str

    def __post_init__(self) -> None:
        msg: list[str] = []
        if not is_decimal(self.segment_start):
            msg.append(f"Segment Start Value: {self.segment_start} | Type: {type(self.segment_start)}")
        if not is_decimal(self.segment_end):
            msg.append(f"Segment End Value: {self.segment_end} | Type: {type(self.segment_start)}")
        if msg:
            title = (
                f"The resource with the ID: '{self.res_id}' expects a float or integer for segment bounds. "
                f"The following places have an unexpected type:"
            )
            wrng = f"{title}{LIST_SEPARATOR}{LIST_SEPARATOR.join(msg)}"
            warnings.warn(DspToolsUserWarning(wrng))


@dataclass
class VideoSegmentResource:
    res_id: str
    label: str
    segment_of: str
    segment_bounds: SegmentBounds
    title: str | None = None
    comments: list[str] = field(default_factory=list)
    descriptions: list[str] = field(default_factory=list)
    keywords: list[str] = field(default_factory=list)
    relates_to: list[str] = field(default_factory=list)
    permissions: str = "res-default"
    migration_metadata: MigrationMetadata | None = None

    def new(
        self,
        res_id: str,
        label: str,
        segment_of: str,
        segment_start: float,
        segment_end: float,
        title: str | None = None,
        permissions: str = "res-default",
    ) -> VideoSegmentResource:
        return VideoSegmentResource(
            res_id=res_id,
            label=label,
            segment_of=segment_of,
            segment_bounds=SegmentBounds(segment_start, segment_end, res_id),
            title=title,
            permissions=permissions,
        )

    def add_title(self, title: str) -> VideoSegmentResource:
        if self.title:
            _warn_value_exists(self.title, title, "title", self.res_id)
        self.title = title
        return self

    def add_comment(self, comment: str) -> VideoSegmentResource:
        self.comments.append(comment)
        return self

    def add_comments(self, comments: list[str]) -> VideoSegmentResource:
        self.comments.extend(comments)
        return self

    def add_description(self, description: str) -> VideoSegmentResource:
        self.descriptions.append(description)
        return self

    def add_descriptions(self, descriptions: list[str]) -> VideoSegmentResource:
        self.descriptions.extend(descriptions)
        return self

    def add_keyword(self, keywords: str) -> VideoSegmentResource:
        self.keywords.append(keywords)
        return self

    def add_keywords(self, keywords: list[str]) -> VideoSegmentResource:
        self.keywords.extend(keywords)
        return self

    def add_relates_to(self, relates_to: str) -> VideoSegmentResource:
        self.relates_to.append(relates_to)
        return self

    def add_relates_to_multiple(self, relates_to: list[str]) -> VideoSegmentResource:
        self.relates_to.extend(relates_to)
        return self

    def add_migration_metadata(
        self, creation_date: str | None, iri: str | None = None, ark: str | None = None
    ) -> VideoSegmentResource:
        if self.migration_metadata:
            raise InputError(
                f"The resource with the ID '{self.res_id}' already contains migration metadata, "
                f"no new data can be added."
            )
        self.migration_metadata = MigrationMetadata(creation_date=creation_date, iri=iri, ark=ark, res_id=self.res_id)
        return self

    def serialise(self) -> etree._Element:
        self._check_for_and_convert_unexpected_input()
        res_ele = self._serialise_resource_element()
        res_ele.extend(_serialise_segment_children(self))
        return res_ele

    def _check_for_and_convert_unexpected_input(self) -> None:
        self.comments = _transform_unexpected_input(self.comments, "comments", self.res_id)
        self.descriptions = _transform_unexpected_input(self.descriptions, "descriptions", self.res_id)
        self.keywords = _transform_unexpected_input(self.keywords, "keywords", self.res_id)
        self.relates_to = _transform_unexpected_input(self.relates_to, "relates_to", self.res_id)
        _validate_segment(self)

    def _serialise_resource_element(self) -> etree._Element:
        attribs = {"label": self.label, "id": self.res_id}
        if self.permissions:
            attribs["permissions"] = self.permissions
        return etree.Element(f"{DASCH_SCHEMA}video-segment", attrib=attribs, nsmap=XML_NAMESPACE_MAP)


@dataclass
class AudioSegmentResource:
    res_id: str
    label: str
    segment_of: str
    segment_bounds: SegmentBounds
    title: str | None = None
    comments: list[str] = field(default_factory=list)
    descriptions: list[str] = field(default_factory=list)
    keywords: list[str] = field(default_factory=list)
    relates_to: list[str] = field(default_factory=list)
    permissions: str = "res-default"
    migration_metadata: MigrationMetadata | None = None

    def new(
        self,
        res_id: str,
        label: str,
        segment_of: str,
        segment_start: float,
        segment_end: float,
        title: str | None = None,
        permissions: str = "res-default",
    ) -> AudioSegmentResource:
        return AudioSegmentResource(
            res_id=res_id,
            label=label,
            segment_of=segment_of,
            segment_bounds=SegmentBounds(segment_start, segment_end, res_id),
            title=title,
            permissions=permissions,
        )

    def add_title(self, title: str) -> AudioSegmentResource:
        if self.title:
            _warn_value_exists(self.title, title, "title", self.res_id)
        self.title = title
        return self

    def add_comment(self, comment: str) -> AudioSegmentResource:
        self.comments.append(comment)
        return self

    def add_comments(self, comments: list[str]) -> AudioSegmentResource:
        self.comments.extend(comments)
        return self

    def add_description(self, description: str) -> AudioSegmentResource:
        self.descriptions.append(description)
        return self

    def add_descriptions(self, descriptions: list[str]) -> AudioSegmentResource:
        self.descriptions.extend(descriptions)
        return self

    def add_keyword(self, keywords: str) -> AudioSegmentResource:
        self.keywords.append(keywords)
        return self

    def add_keywords(self, keywords: list[str]) -> AudioSegmentResource:
        self.keywords.extend(keywords)
        return self

    def add_relates_to(self, relates_to: str) -> AudioSegmentResource:
        self.relates_to.append(relates_to)
        return self

    def add_relates_to_multiple(self, relates_to: list[str]) -> AudioSegmentResource:
        self.relates_to.extend(relates_to)
        return self

    def add_migration_metadata(
        self, creation_date: str | None, iri: str | None = None, ark: str | None = None
    ) -> AudioSegmentResource:
        if self.migration_metadata:
            raise InputError(
                f"The resource with the ID '{self.res_id}' already contains migration metadata, "
                f"no new data can be added."
            )
        self.migration_metadata = MigrationMetadata(creation_date=creation_date, iri=iri, ark=ark, res_id=self.res_id)
        return self

    def serialise(self) -> etree._Element:
        self._check_for_and_convert_unexpected_input()
        res_ele = self._serialise_resource_element()
        res_ele.extend(_serialise_segment_children(self))
        return res_ele

    def _check_for_and_convert_unexpected_input(self) -> None:
        self.comments = _transform_unexpected_input(self.comments, "comments", self.res_id)
        self.descriptions = _transform_unexpected_input(self.descriptions, "descriptions", self.res_id)
        self.keywords = _transform_unexpected_input(self.keywords, "keywords", self.res_id)
        self.relates_to = _transform_unexpected_input(self.relates_to, "relates_to", self.res_id)
        _validate_segment(self)

    def _serialise_resource_element(self) -> etree._Element:
        attribs = {"label": self.label, "id": self.res_id}
        if self.permissions:
            attribs["permissions"] = self.permissions
        return etree.Element(f"{DASCH_SCHEMA}audio-segment", attrib=attribs, nsmap=XML_NAMESPACE_MAP)


def _check_strings(string_to_check: str, res_id: str, field_name: str) -> None:
    if not is_string_like(string_to_check):
        msg = (
            f"The resource with the ID '{res_id}' has an invalid string at the following location:\n"
            f"Field: {field_name} | Value: {string_to_check}"
        )
        warnings.warn(DspToolsUserWarning(msg))


def _serialise_has_comment(comments: list[str], res_id: str) -> etree._Element:
    cmts = [Richtext(value=x, prop_name="hasComment", resource_id=res_id) for x in comments]
    cmt_prop = cmts[0].make_prop()
    cmt_prop.extend([cmt.make_element() for cmt in cmts])
    return cmt_prop


def _validate_segment(segment: AudioSegmentResource | VideoSegmentResource) -> None:
    problems = []
    if not is_string_like(segment.res_id):
        problems.append(f"Field: Resource ID | Value: {segment.res_id}")
    if not is_string_like(segment.label):
        problems.append(f"Field: label | Value: {segment.label}")
    if not is_string_like(segment.segment_of):
        problems.append(f"Field: segment_of | Value: {segment.segment_of}")
    if segment.title and not is_string_like(segment.title):
        problems.append(f"Field: title | Value: {segment.title}")
    if fails := [x for x in segment.comments if not is_string_like(x)]:
        problems.extend([f"Field: comment | Value: {x}" for x in fails])
    if fails := [x for x in segment.descriptions if not is_string_like(x)]:
        problems.extend([f"Field: description | Value: {x}" for x in fails])
    if fails := [x for x in segment.keywords if not is_string_like(x)]:
        problems.extend([f"Field: keywords | Value: {x}" for x in fails])
    if fails := [x for x in segment.relates_to if not is_string_like(x)]:
        problems.extend([f"Field: relates_to | Value: {x}" for x in fails])
    if problems:
        msg = f"The resource with the ID '{segment.res_id}' has the following problem(s):{'\n- '.join(problems)}"
        warnings.warn(DspToolsUserWarning(msg))


def _serialise_segment_children(segment: AudioSegmentResource | VideoSegmentResource) -> list[etree._Element]:
    segment_elements = []
    segment_of = etree.Element(f"{DASCH_SCHEMA}isSegmentOf", nsmap=XML_NAMESPACE_MAP)
    segment_of.text = segment.segment_of
    segment_elements.append(segment_of)
    segment_elements.append(
        etree.Element(
            f"{DASCH_SCHEMA}hasSegmentBounds",
            attrib={"start": str(segment.segment_bounds.segment_start), "end": str(segment.segment_bounds.segment_end)},
            nsmap=XML_NAMESPACE_MAP,
        )
    )
    if segment.title:
        segment_elements.append(_make_element_with_text("hasTitle", segment.title))
    segment_elements.extend([_make_element_with_text("hasComment", x) for x in segment.comments])
    segment_elements.extend([_make_element_with_text("hasDescription", x) for x in segment.descriptions])
    segment_elements.extend([_make_element_with_text("hasKeyword", x) for x in segment.keywords])
    segment_elements.extend([_make_element_with_text("relatesTo", x) for x in segment.relates_to])
    return segment_elements


def _make_element_with_text(tag_name: str, text_content: str) -> etree._Element:
    ele = etree.Element(f"{DASCH_SCHEMA}{tag_name}", nsmap=XML_NAMESPACE_MAP)
    ele.text = text_content
    return ele


def _transform_unexpected_input(value: Any, prop_name: str, res_id: str) -> list[str]:
    match value:
        case list():
            return value
        case set() | tuple():
            return list(value)
        case str():
            return [value]
        case _:
            _warn_unexpected_value(value, prop_name, res_id)
            return [str(value)]


def _warn_unexpected_value(value: Any, prop_name: str, res_id: str | None) -> None:
    msg = (
        f"The resource: {res_id} should have a list of strings for the field '{prop_name}'. "
        f"Your input: '{value}' is of type {type(value)}."
    )
    warnings.warn(DspToolsUserWarning(msg))


def _warn_value_exists(old_value: Any, new_value: Any, value_field: str, res_id: str | None) -> None:
    """Emits a warning if a values is not in the expected format."""
    msg = (
        f"The resource with the ID '{res_id}' already has a value in the field '{value_field}'. "
        f"The old value '{old_value}' is being replace with '{new_value}'."
    )
    warnings.warn(DspToolsUserWarning(msg))



from rdflib import RDF
from rdflib import SH
from rdflib import Graph
from rdflib.term import Node

from dsp_tools.commands.xml_validate.models.input_problems import AllProblems
from dsp_tools.commands.xml_validate.models.input_problems import InputProblem
from dsp_tools.commands.xml_validate.models.input_problems import MaxCardinalityViolation
from dsp_tools.commands.xml_validate.models.input_problems import MinCardinalityViolation
from dsp_tools.commands.xml_validate.models.input_problems import NonExistentCardinalityViolation
from dsp_tools.commands.xml_validate.models.input_problems import UnexpectedResults
from dsp_tools.commands.xml_validate.models.validation import UnexpectedComponent
from dsp_tools.commands.xml_validate.models.validation import ValidationResult


def reformat_validation_graph(results_graph: Graph, data_graph: Graph) -> AllProblems:
    """
    Reformats the validation result from an RDF graph into class instances
    that are used to communicate the problems with the user.

    Args:
        results_graph: Contains the possible validation errors
        data_graph: Graph with the data

    Returns:
        All Problems
    """
    reformatted_results = _reformat_result_graph(results_graph, data_graph)
    input_problems, unexpected_components = _transform_violations_into_input_problems(reformatted_results)
    unexpected = UnexpectedResults(unexpected_components) if unexpected_components else None
    return AllProblems(input_problems, unexpected)


def _reformat_result_graph(results_graph: Graph, data_graph: Graph) -> list[ValidationResult]:
    violations = results_graph.subjects(RDF.type, SH.ValidationResult)
    return [_extract_one_violation(x, results_graph, data_graph) for x in violations]


def _extract_one_violation(bn: Node, results_graph: Graph, data_graph: Graph) -> ValidationResult:
    focus_nd = next(results_graph.objects(bn, SH.focusNode))
    res_type = next(data_graph.objects(focus_nd, RDF.type))
    path = next(results_graph.objects(bn, SH.resultPath))
    component = next(results_graph.objects(bn, SH.sourceConstraintComponent))
    msg = str(next(results_graph.objects(bn, SH.resultMessage)))
    return ValidationResult(
        source_constraint_component=component,
        res_iri=focus_nd,
        res_class=res_type,
        property=path,
        results_message=msg,
    )


def _transform_violations_into_input_problems(
    violations: list[ValidationResult],
) -> tuple[list[InputProblem], list[UnexpectedComponent]]:
    input_problems: list[InputProblem] = []
    unexpected_components: list[UnexpectedComponent] = []
    for violation in violations:
        problem = _reformat_one_violation(violation)
        if isinstance(problem, UnexpectedComponent):
            unexpected_components.append(problem)
        else:
            input_problems.append(problem)
    return input_problems, unexpected_components


def _reformat_one_violation(violation: ValidationResult) -> InputProblem | UnexpectedComponent:
    subject_id = _reformat_data_iri(str(violation.res_iri))
    prop_name = _reformat_onto_iri(str(violation.property))
    res_type = _reformat_onto_iri(str(violation.res_class))
    match violation.source_constraint_component:
        case SH.MaxCountConstraintComponent:
            return MaxCardinalityViolation(subject_id, res_type, prop_name, violation.results_message)
        case SH.MinCountConstraintComponent:
            return MinCardinalityViolation(subject_id, res_type, prop_name, violation.results_message)
        case SH.ClosedConstraintComponent:
            return NonExistentCardinalityViolation(subject_id, res_type, prop_name)
        case _:
            return UnexpectedComponent(str(violation.source_constraint_component))


def _reformat_onto_iri(prop: str) -> str:
    onto = prop.split("/")[-2]
    return f'{onto}:{prop.split("#")[-1]}'


def _reformat_data_iri(iri: str) -> str:
    return iri.replace("http://data/", "")



from __future__ import annotations

import warnings
from collections import defaultdict
from dataclasses import dataclass
from dataclasses import field
from typing import Any

import pandas as pd
from lxml import etree

from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.models.exceptions import InputError
from dsp_tools.xmllib.models.file_values import AbstractFileValue
from dsp_tools.xmllib.models.file_values import FileValue
from dsp_tools.xmllib.models.file_values import IIIFUri
from dsp_tools.xmllib.models.migration_metadata import MigrationMetadata
from dsp_tools.xmllib.models.values import BooleanValue
from dsp_tools.xmllib.models.values import ColorValue
from dsp_tools.xmllib.models.values import DateValue
from dsp_tools.xmllib.models.values import DecimalValue
from dsp_tools.xmllib.models.values import GeonameValue
from dsp_tools.xmllib.models.values import IntValue
from dsp_tools.xmllib.models.values import LinkValue
from dsp_tools.xmllib.models.values import ListValue
from dsp_tools.xmllib.models.values import Richtext
from dsp_tools.xmllib.models.values import SimpleText
from dsp_tools.xmllib.models.values import TimeValue
from dsp_tools.xmllib.models.values import UriValue
from dsp_tools.xmllib.models.values import Value
from dsp_tools.xmllib.value_checkers import is_string_like

XML_NAMESPACE_MAP = {None: "https://dasch.swiss/schema", "xsi": "http://www.w3.org/2001/XMLSchema-instance"}
DASCH_SCHEMA = "{https://dasch.swiss/schema}"

LIST_SEPARATOR = "\n    - "


@dataclass
class Resource:
    res_id: str
    restype: str
    label: str
    values: list[Value] = field(default_factory=list)
    permissions: str = "res-default"
    file_value: AbstractFileValue | None = None
    migration_metadata: MigrationMetadata | None = None

    def __post_init__(self) -> None:
        msg = []
        if not is_string_like(str(self.label)):
            msg.append(f"Label '{self.label}'")
        if not is_string_like(str(self.res_id)):
            msg.append(f"Resource ID '{self.res_id}'")
        if not is_string_like(str(self.restype)):
            msg.append(f"Resource Type '{self.restype}'")
        if msg:
            out_msg = (
                f"The Resource with the ID '{self.res_id}' should have strings in the following field(s), "
                f"the input is not a valid string.:{LIST_SEPARATOR}{LIST_SEPARATOR.join(msg)}"
            )
            warnings.warn(DspToolsUserWarning(out_msg))

    def new(self, res_id: str, restype: str, label: str, permissions: str = "res-default") -> Resource:
        return Resource(
            res_id=res_id,
            restype=restype,
            label=label,
            permissions=permissions,
        )

    def serialise(self) -> etree._Element:
        res_ele = self._serialise_resource_element()
        if self.file_value:
            res_ele.append(self.file_value.serialise())
        res_ele.extend(self._serialise_values())
        return res_ele

    def _serialise_resource_element(self) -> etree._Element:
        attribs = {"label": self.label, "restype": self.restype, "id": self.res_id}
        if self.permissions:
            attribs["permissions"] = self.permissions
        return etree.Element(f"{DASCH_SCHEMA}resource", attrib=attribs, nsmap=XML_NAMESPACE_MAP)

    def _serialise_values(self) -> list[etree._Element]:
        grouped = defaultdict(list)
        for val in self.values:
            grouped[val.prop_name].append(val)
        return [self._combine_values(prop_values) for prop_values in grouped.values()]

    def _combine_values(self, prop_values: list[Value]) -> etree._Element:
        prop_ = prop_values[0].make_prop()
        prop_eles = [x.make_element() for x in prop_values]
        prop_.extend(prop_eles)
        return prop_

    #######################
    # BooleanValue
    #######################

    def add_bool(
        self, value: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.append(BooleanValue(value, prop_name, permissions, comment, self.res_id))
        return self

    def add_bools(
        self, values: list[Any], prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.extend([BooleanValue(v, prop_name, permissions, comment, self.res_id) for v in values])
        return self

    def add_bool_optional(
        self, value: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        if not pd.isna(value):
            self.values.append(BooleanValue(value, prop_name, permissions, comment, self.res_id))
        return self

    #######################
    # ColorValue
    #######################

    def add_color(
        self, value: int | str, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.append(ColorValue(value, prop_name, permissions, comment, self.res_id))
        return self

    def add_colors(
        self, values: list[int | str], prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.extend([ColorValue(v, prop_name, permissions, comment, self.res_id) for v in values])
        return self

    def add_color_optional(
        self, value: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        if not pd.isna(value):
            self.values.append(ColorValue(value, prop_name, permissions, comment, self.res_id))
        return self

    #######################
    # DateValue
    #######################

    def add_date(
        self, value: str, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.append(DateValue(value, prop_name, permissions, comment, self.res_id))
        return self

    def add_dates(
        self, values: list[str], prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.extend([DateValue(v, prop_name, permissions, comment, self.res_id) for v in values])
        return self

    def add_date_optional(
        self, value: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        if not pd.isna(value):
            self.values.append(DateValue(value, prop_name, permissions, comment, self.res_id))
        return self

    #######################
    # DecimalValue
    #######################

    def add_decimal(
        self, value: float | str, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.append(DecimalValue(value, prop_name, permissions, comment, self.res_id))
        return self

    def add_decimals(
        self, values: list[float | str], prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.extend([DecimalValue(v, prop_name, permissions, comment, self.res_id) for v in values])
        return self

    def add_decimal_optional(
        self, value: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        if not pd.isna(value):
            self.values.append(DecimalValue(value, prop_name, permissions, comment, self.res_id))
        return self

    #######################
    # GeonameValue
    #######################

    def add_geoname(
        self, value: int | str, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.append(GeonameValue(value, prop_name, permissions, comment, self.res_id))
        return self

    def add_geonames(
        self, values: list[int | str], prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.extend([GeonameValue(v, prop_name, permissions, comment, self.res_id) for v in values])
        return self

    def add_geoname_optional(
        self, value: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        if not pd.isna(value):
            self.values.append(GeonameValue(value, prop_name, permissions, comment, self.res_id))
        return self

    #######################
    # IntValue
    #######################

    def add_integer(
        self, value: int | str, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.append(IntValue(value, prop_name, permissions, comment, self.res_id))
        return self

    def add_integers(
        self, values: list[int | str], prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.extend([IntValue(v, prop_name, permissions, comment, self.res_id) for v in values])
        return self

    def add_integer_optional(
        self, value: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        if not pd.isna(value):
            self.values.append(IntValue(value, prop_name, permissions, comment, self.res_id))
        return self

    #######################
    # LinkValue
    #######################

    def add_link(
        self, value: str, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.append(LinkValue(value, prop_name, permissions, comment, self.res_id))
        return self

    def add_links(
        self, values: list[str], prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.extend([LinkValue(v, prop_name, permissions, comment, self.res_id) for v in values])
        return self

    def add_link_optional(
        self, value: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        if not pd.isna(value):
            self.values.append(LinkValue(value, prop_name, permissions, comment, self.res_id))
        return self

    #######################
    # ListValue
    #######################

    def add_list(
        self, value: Any, list_name: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.append(ListValue(value, list_name, prop_name, permissions, comment, self.res_id))
        return self

    def add_lists(
        self,
        values: list[Any],
        list_name: Any,
        prop_name: str,
        permissions: str | None = None,
        comment: str | None = None,
    ) -> Resource:
        self.values.extend([ListValue(v, list_name, prop_name, permissions, comment, self.res_id) for v in values])
        return self

    def add_list_optional(
        self, value: Any, list_name: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        if not pd.isna(value):
            self.values.append(ListValue(value, list_name, prop_name, permissions, comment, self.res_id))
        return self

    #######################
    # TextValue: SimpleText
    #######################

    def add_simpletext(
        self, value: str, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.append(SimpleText(value, prop_name, permissions, comment, self.res_id))
        return self

    def add_simpletexts(
        self, values: list[str], prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.extend([SimpleText(v, prop_name, permissions, comment, self.res_id) for v in values])
        return self

    def add_simpletext_optional(
        self, value: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        if not pd.isna(value):
            self.values.append(SimpleText(value, prop_name, permissions, comment, self.res_id))
        return self

    #######################
    # TextValue: Richtext
    #######################

    def add_richtext(
        self, value: str, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.append(Richtext(value, prop_name, permissions, comment, self.res_id))
        return self

    def add_richtexts(
        self, values: list[str], prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.extend([Richtext(v, prop_name, permissions, comment, self.res_id) for v in values])
        return self

    def add_richtext_optional(
        self, value: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        if not pd.isna(value):
            self.values.append(Richtext(value, prop_name, permissions, comment, self.res_id))
        return self

    #######################
    # TimeValue
    #######################

    def add_time(
        self, value: str, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.append(TimeValue(value, prop_name, permissions, comment, self.res_id))
        return self

    def add_times(
        self, values: list[str], prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.extend([TimeValue(v, prop_name, permissions, comment, self.res_id) for v in values])
        return self

    def add_time_optional(
        self, value: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        if not pd.isna(value):
            self.values.append(TimeValue(value, prop_name, permissions, comment, self.res_id))
        return self

    #######################
    # UriValue
    #######################

    def add_uri(
        self, value: str, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.append(UriValue(value, prop_name, permissions, comment, self.res_id))
        return self

    def add_uris(
        self, values: list[str], prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        self.values.extend([UriValue(v, prop_name, permissions, comment, self.res_id) for v in values])
        return self

    def add_uri_optional(
        self, value: Any, prop_name: str, permissions: str | None = None, comment: str | None = None
    ) -> Resource:
        if not pd.isna(value):
            self.values.append(UriValue(value, prop_name, permissions, comment, self.res_id))
        return self

    #######################
    # AbstractFileValue
    #######################

    def add_file(self, filename: str, permissions: str | None = None, comment: str | None = None) -> Resource:
        if self.file_value:
            raise InputError(
                f"The resource with the ID '{self.res_id}' already contains a file with the name: "
                f"'{self.file_value.value}'.\n"
                f"The new file with the name '{filename}' cannot be added."
            )
        self.file_value = FileValue(filename, permissions, comment, self.res_id)
        return self

    def add_iiif_uri(self, iiif_uri: str, permissions: str | None = None, comment: str | None = None) -> Resource:
        if self.file_value:
            raise InputError(
                f"The resource with the ID '{self.res_id}' already contains a file with the name: "
                f"'{self.file_value.value}'.\n"
                f"The new file with the name '{iiif_uri}' cannot be added."
            )
        self.file_value = IIIFUri(iiif_uri, permissions, comment, self.res_id)
        return self

    #######################
    # Migration Metadata
    #######################

    def add_migration_metadata(
        self, creation_date: str | None, iri: str | None = None, ark: str | None = None
    ) -> Resource:
        if self.migration_metadata:
            raise InputError(
                f"The resource with the ID '{self.res_id}' already contains migration metadata, "
                f"no new data can be added."
            )
        self.migration_metadata = MigrationMetadata(creation_date=creation_date, iri=iri, ark=ark, res_id=self.res_id)
        return self



import pytest
from rdflib import Graph
from rdflib import Namespace

ONTO = Namespace("http://0.0.0.0:3333/ontology/9999/onto/v2#")
API_SHAPES = Namespace("http://api.knora.org/ontology/knora-api/shapes/v2#")
PREFIXES = """
@prefix knora-api: <http://api.knora.org/ontology/knora-api/v2#> .
@prefix onto: <http://0.0.0.0:3333/ontology/9999/onto/v2#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix salsah-gui: <http://api.knora.org/ontology/salsah-gui/v2#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
"""


@pytest.fixture
def link_prop_card_1() -> Graph:
    ttl = f"""{PREFIXES}
    onto:ClassMixedCard a owl:Class ;
    rdfs:label "Resource with all cardinality options" ;
    knora-api:canBeInstantiated true ;
    knora-api:isResourceClass true ;
    rdfs:subClassOf [ 
            a owl:Restriction ;
            salsah-gui:guiOrder 1 ;
            owl:cardinality 1 ;
            owl:onProperty onto:testHasLinkToCardOneResourceValue 
            ] ,
            [ a owl:Restriction ;
            salsah-gui:guiOrder 1 ;
            owl:cardinality 1 ;
            owl:onProperty onto:testHasLinkToCardOneResource ] .
            
    onto:testHasLinkToCardOneResource a owl:ObjectProperty ;
        rdfs:label "Super-class" ;
        knora-api:isEditable true ;
        knora-api:isLinkProperty true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType onto:CardOneResource ;
        salsah-gui:guiElement salsah-gui:Searchbox ;
        rdfs:subPropertyOf knora-api:hasLinkTo .
    
    onto:testHasLinkToCardOneResourceValue a owl:ObjectProperty ;
        rdfs:label "Super-class" ;
        knora-api:isEditable true ;
        knora-api:isLinkValueProperty true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType knora-api:LinkValue ;
        salsah-gui:guiElement salsah-gui:Searchbox ;
        rdfs:subPropertyOf knora-api:hasLinkToValue .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def link_prop_card_01() -> Graph:
    ttl = f"""{PREFIXES}
    onto:ClassMixedCard a owl:Class ;
    rdfs:label "Resource with all cardinality options" ;
    knora-api:canBeInstantiated true ;
    knora-api:isResourceClass true ;
    rdfs:subClassOf [ 
            a owl:Restriction ;
            salsah-gui:guiOrder 1 ;
            owl:maxCardinality 1 ;
            owl:onProperty onto:testHasLinkToCardOneResourceValue 
            ] ,
            [ a owl:Restriction ;
            salsah-gui:guiOrder 1 ;
            owl:maxCardinality 1 ;
            owl:onProperty onto:testHasLinkToCardOneResource ] .

    onto:testHasLinkToCardOneResource a owl:ObjectProperty ;
        rdfs:label "Super-class" ;
        knora-api:isEditable true ;
        knora-api:isLinkProperty true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType onto:CardOneResource ;
        salsah-gui:guiElement salsah-gui:Searchbox ;
        rdfs:subPropertyOf knora-api:hasLinkTo .

    onto:testHasLinkToCardOneResourceValue a owl:ObjectProperty ;
        rdfs:label "Super-class" ;
        knora-api:isEditable true ;
        knora-api:isLinkValueProperty true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType knora-api:LinkValue ;
        salsah-gui:guiElement salsah-gui:Searchbox ;
        rdfs:subPropertyOf knora-api:hasLinkToValue .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g



from typing import cast

import pytest
from lxml import etree
from rdflib import RDF
from rdflib import BNode
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import IIIFUriInfo
from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLValue
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import BitstreamInfo
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import XMLResource
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.permission import PermissionValue
from dsp_tools.commands.xmlupload.models.serialise.jsonld_serialiser import serialise_property_graph
from dsp_tools.commands.xmlupload.resource_create_client import KNORA_API
from dsp_tools.commands.xmlupload.resource_create_client import _make_bitstream_file_value
from dsp_tools.commands.xmlupload.resource_create_client import _make_boolean_value
from dsp_tools.commands.xmlupload.resource_create_client import _make_iiif_uri_value
from dsp_tools.commands.xmlupload.resource_create_client import _to_boolean
from dsp_tools.models.exceptions import BaseError
from dsp_tools.models.exceptions import PermissionNotExistsError


class TestMakeBitstreamFileValue:
    """Tests the _make_bitstream_file_value function."""

    def test_zip(self) -> None:
        info = BitstreamInfo("a/b/test.zip", "00001.zip")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasArchiveFileValue": {
                "@type": "knora-api:ArchiveFileValue",
                "knora-api:fileValueHasFilename": "00001.zip",
            }
        }
        assert value == expected

    def test_tar(self) -> None:
        info = BitstreamInfo("a/b/test.tar", "00001.tar")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasArchiveFileValue": {
                "@type": "knora-api:ArchiveFileValue",
                "knora-api:fileValueHasFilename": "00001.tar",
            }
        }
        assert value == expected

    def test_gz(self) -> None:
        info = BitstreamInfo("a/b/test.gz", "00001.gz")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasArchiveFileValue": {
                "@type": "knora-api:ArchiveFileValue",
                "knora-api:fileValueHasFilename": "00001.gz",
            }
        }
        assert value == expected

    def test_z(self) -> None:
        info = BitstreamInfo("a/b/test.z", "00001.z")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasArchiveFileValue": {
                "@type": "knora-api:ArchiveFileValue",
                "knora-api:fileValueHasFilename": "00001.z",
            }
        }
        assert value == expected

    def test_tar_gz(self) -> None:
        info = BitstreamInfo("a/b/test.tar.gz", "00001.tar.gz")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasArchiveFileValue": {
                "@type": "knora-api:ArchiveFileValue",
                "knora-api:fileValueHasFilename": "00001.tar.gz",
            }
        }
        assert value == expected

    def test_tgz(self) -> None:
        info = BitstreamInfo("a/b/test.tgz", "00001.tgz")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasArchiveFileValue": {
                "@type": "knora-api:ArchiveFileValue",
                "knora-api:fileValueHasFilename": "00001.tgz",
            }
        }
        assert value == expected

    def test_gzip(self) -> None:
        info = BitstreamInfo("a/b/test.gzip", "00001.gzip")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasArchiveFileValue": {
                "@type": "knora-api:ArchiveFileValue",
                "knora-api:fileValueHasFilename": "00001.gzip",
            }
        }
        assert value == expected

    def test_7z(self) -> None:
        info = BitstreamInfo("a/b/test.7z", "00001.7z")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasArchiveFileValue": {
                "@type": "knora-api:ArchiveFileValue",
                "knora-api:fileValueHasFilename": "00001.7z",
            }
        }
        assert value == expected

    def test_mp3(self) -> None:
        info = BitstreamInfo("a/b/test.mp3", "00001.mp3")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasAudioFileValue": {
                "@type": "knora-api:AudioFileValue",
                "knora-api:fileValueHasFilename": "00001.mp3",
            }
        }
        assert value == expected

    def test_wav(self) -> None:
        info = BitstreamInfo("a/b/test.wav", "00001.wav")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasAudioFileValue": {
                "@type": "knora-api:AudioFileValue",
                "knora-api:fileValueHasFilename": "00001.wav",
            }
        }
        assert value == expected

    def test_pdf(self) -> None:
        info = BitstreamInfo("a/b/test.pdf", "00001.pdf")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasDocumentFileValue": {
                "@type": "knora-api:DocumentFileValue",
                "knora-api:fileValueHasFilename": "00001.pdf",
            }
        }
        assert value == expected

    def test_doc(self) -> None:
        info = BitstreamInfo("a/b/test.doc", "00001.doc")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasDocumentFileValue": {
                "@type": "knora-api:DocumentFileValue",
                "knora-api:fileValueHasFilename": "00001.doc",
            }
        }
        assert value == expected

    def test_docx(self) -> None:
        info = BitstreamInfo("a/b/test.docx", "00001.docx")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasDocumentFileValue": {
                "@type": "knora-api:DocumentFileValue",
                "knora-api:fileValueHasFilename": "00001.docx",
            }
        }
        assert value == expected

    def test_xls(self) -> None:
        info = BitstreamInfo("a/b/test.xls", "00001.xls")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasDocumentFileValue": {
                "@type": "knora-api:DocumentFileValue",
                "knora-api:fileValueHasFilename": "00001.xls",
            }
        }
        assert value == expected

    def test_xlsx(self) -> None:
        info = BitstreamInfo("a/b/test.xlsx", "00001.xlsx")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasDocumentFileValue": {
                "@type": "knora-api:DocumentFileValue",
                "knora-api:fileValueHasFilename": "00001.xlsx",
            }
        }
        assert value == expected

    def test_ppt(self) -> None:
        info = BitstreamInfo("a/b/test.ppt", "00001.ppt")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasDocumentFileValue": {
                "@type": "knora-api:DocumentFileValue",
                "knora-api:fileValueHasFilename": "00001.ppt",
            }
        }
        assert value == expected

    def test_pptx(self) -> None:
        info = BitstreamInfo("a/b/test.pptx", "00001.pptx")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasDocumentFileValue": {
                "@type": "knora-api:DocumentFileValue",
                "knora-api:fileValueHasFilename": "00001.pptx",
            }
        }
        assert value == expected

    def test_mp4(self) -> None:
        info = BitstreamInfo("a/b/test.mp4", "00001.mp4")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasMovingImageFileValue": {
                "@type": "knora-api:MovingImageFileValue",
                "knora-api:fileValueHasFilename": "00001.mp4",
            }
        }
        assert value == expected

    def test_jpg(self) -> None:
        info = BitstreamInfo("a/b/test.jpg", "00001.jp2")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasStillImageFileValue": {
                "@type": "knora-api:StillImageFileValue",
                "knora-api:fileValueHasFilename": "00001.jp2",
            }
        }
        assert value == expected

    def test_jpeg(self) -> None:
        info = BitstreamInfo("a/b/test.jpeg", "00001.jp2")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasStillImageFileValue": {
                "@type": "knora-api:StillImageFileValue",
                "knora-api:fileValueHasFilename": "00001.jp2",
            }
        }
        assert value == expected

    def test_jp2(self) -> None:
        info = BitstreamInfo("a/b/test.jp2", "00001.jp2")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasStillImageFileValue": {
                "@type": "knora-api:StillImageFileValue",
                "knora-api:fileValueHasFilename": "00001.jp2",
            }
        }
        assert value == expected

    def test_png(self) -> None:
        info = BitstreamInfo("a/b/test.png", "00001.jp2")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasStillImageFileValue": {
                "@type": "knora-api:StillImageFileValue",
                "knora-api:fileValueHasFilename": "00001.jp2",
            }
        }
        assert value == expected

    def test_tif(self) -> None:
        info = BitstreamInfo("a/b/test.tif", "00001.jp2")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasStillImageFileValue": {
                "@type": "knora-api:StillImageFileValue",
                "knora-api:fileValueHasFilename": "00001.jp2",
            }
        }
        assert value == expected

    def test_tiff(self) -> None:
        info = BitstreamInfo("a/b/test.tiff", "00001.jp2")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasStillImageFileValue": {
                "@type": "knora-api:StillImageFileValue",
                "knora-api:fileValueHasFilename": "00001.jp2",
            }
        }
        assert value == expected

    def test_odd(self) -> None:
        info = BitstreamInfo("a/b/test.odd", "00001.odd")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasTextFileValue": {
                "@type": "knora-api:TextFileValue",
                "knora-api:fileValueHasFilename": "00001.odd",
            }
        }
        assert value == expected

    def test_rng(self) -> None:
        info = BitstreamInfo("a/b/test.rng", "00001.rng")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasTextFileValue": {
                "@type": "knora-api:TextFileValue",
                "knora-api:fileValueHasFilename": "00001.rng",
            }
        }
        assert value == expected

    def test_txt(self) -> None:
        info = BitstreamInfo("a/b/test.txt", "00001.txt")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasTextFileValue": {
                "@type": "knora-api:TextFileValue",
                "knora-api:fileValueHasFilename": "00001.txt",
            }
        }
        assert value == expected

    def test_xml(self) -> None:
        info = BitstreamInfo("a/b/test.xml", "00001.xml")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasTextFileValue": {
                "@type": "knora-api:TextFileValue",
                "knora-api:fileValueHasFilename": "00001.xml",
            }
        }
        assert value == expected

    def test_xsd(self) -> None:
        info = BitstreamInfo("a/b/test.xsd", "00001.xsd")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasTextFileValue": {
                "@type": "knora-api:TextFileValue",
                "knora-api:fileValueHasFilename": "00001.xsd",
            }
        }
        assert value == expected

    def test_xsl(self) -> None:
        info = BitstreamInfo("a/b/test.xsl", "00001.xsl")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasTextFileValue": {
                "@type": "knora-api:TextFileValue",
                "knora-api:fileValueHasFilename": "00001.xsl",
            }
        }
        assert value == expected

    def test_csv(self) -> None:
        info = BitstreamInfo("a/b/test.csv", "00001.csv")
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasTextFileValue": {
                "@type": "knora-api:TextFileValue",
                "knora-api:fileValueHasFilename": "00001.csv",
            }
        }
        assert value == expected

    def test_with_permissions(self) -> None:
        permission = Permissions({PermissionValue.CR: ["knora-admin:ProjectAdmin"]})
        info = BitstreamInfo("a/b/test.csv", "00001.csv", permission)
        value = _make_bitstream_file_value(info)
        expected = {
            "knora-api:hasTextFileValue": {
                "@type": "knora-api:TextFileValue",
                "knora-api:hasPermissions": "CR knora-admin:ProjectAdmin",
                "knora-api:fileValueHasFilename": "00001.csv",
            }
        }
        assert value == expected


def test_to_boolean() -> None:
    assert _to_boolean("true")
    assert _to_boolean("True")
    assert _to_boolean("1")
    assert _to_boolean(1)
    assert _to_boolean(True)
    assert not _to_boolean("false")
    assert not _to_boolean("False")
    assert not _to_boolean("0")
    assert not _to_boolean(0)
    assert not _to_boolean(False)
    with pytest.raises(BaseError):
        _to_boolean("foo")
    with pytest.raises(BaseError):
        _to_boolean(2)


def test_make_iiif_uri_value_with_permissions() -> None:
    permission = {"prop-default": Permissions({PermissionValue.CR: ["knora-admin:ProjectAdmin"]})}
    xml_str = """
        <resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id">
            <iiif-uri permissions="prop-default">http://example.org/prefix1/abcd1234/full/full/0/native.jpg</iiif-uri>
        </resource>
        """
    xmlresource = XMLResource.from_node(etree.fromstring(xml_str), "foo")
    test_val = cast(IIIFUriInfo, xmlresource.iiif_uri)
    result = _make_iiif_uri_value(test_val, BNode(), permission)
    assert len(result) == 4


def test_make_iiif_uri_value_no_permissions() -> None:
    permission = {"prop-default": Permissions()}
    xml_str = """
        <resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id">
            <iiif-uri>http://example.org/prefix1/abcd1234/full/full/0/native.jpg</iiif-uri>
        </resource>
        """
    xmlresource = XMLResource.from_node(etree.fromstring(xml_str), "foo")
    test_val = cast(IIIFUriInfo, xmlresource.iiif_uri)
    result = _make_iiif_uri_value(test_val, BNode(), permission)
    assert len(result) == 3


def test_make_iiif_uri_value_raises() -> None:
    permission = {"": Permissions()}
    xml_str = """
        <resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id">
            <iiif-uri permissions="prop-default">http://example.org/prefix1/abcd1234/full/full/0/native.jpg</iiif-uri>
        </resource>
        """
    xmlresource = XMLResource.from_node(etree.fromstring(xml_str), "foo")
    test_val = cast(IIIFUriInfo, xmlresource.iiif_uri)
    with pytest.raises(PermissionNotExistsError):
        _make_iiif_uri_value(test_val, BNode(), permission)


def test_make_iiif_uri_value_serialised() -> None:
    xml_str = """
        <resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id">
            <iiif-uri>http://example.org/prefix1/abcd1234/full/full/0/native.jpg</iiif-uri>
        </resource>
        """
    xmlresource = XMLResource.from_node(etree.fromstring(xml_str), "foo")
    test_val = cast(IIIFUriInfo, xmlresource.iiif_uri)
    result = _make_iiif_uri_value(test_val, BNode(), {})
    serialised = serialise_property_graph(
        result, URIRef("http://api.knora.org/ontology/knora-api/v2#hasStillImageFileValue")
    )
    expected = {
        "http://api.knora.org/ontology/knora-api/v2#hasStillImageFileValue": {
            "@type": "http://api.knora.org/ontology/knora-api/v2#StillImageExternalFileValue",
            "http://api.knora.org/ontology/knora-api/v2#fileValueHasExternalUrl": "http://example.org/prefix1/abcd1234/full/full/0/native.jpg",
        }
    }
    assert serialised == expected


def test_make_boolean_value_with_permissions() -> None:
    permissions_lookup = {"prop-default": Permissions({PermissionValue.CR: ["knora-admin:ProjectAdmin"]})}

    xml_str = """
        <resource label="foo_1_label" restype=":foo_1_type" id="foo_1_id">
            <boolean-prop name=":isTrueOrFalse">
                <boolean permissions="prop-default">true</boolean>
            </boolean-prop>
        </resource>
        """
    xmlresource = XMLResource.from_node(etree.fromstring(xml_str), "foo")
    test_val: XMLValue = xmlresource.properties[0].values[0]
    b_node = BNode()
    bool_graph = _make_boolean_value(test_val, b_node, permissions_lookup)

    for triple in bool_graph.triples((b_node, None, None)):
        if triple[1] == URIRef(RDF.type) and triple[2] == URIRef(KNORA_API.BooleanValue):
            continue
        elif triple[1] == URIRef(KNORA_API.booleanValueAsBoolean) and triple[2] == Literal(_to_boolean("true")):
            continue
        elif triple[1] == URIRef(KNORA_API.hasPermissions) and triple[2] == Literal(
            str(permissions_lookup.get(str(test_val.permissions)))
        ):
            continue
        else:
            # unexpected triple
            pytest.fail(f"unexpected triple: {triple}")


if __name__ == "__main__":
    pytest.main([__file__])



from rdflib import Graph

from dsp_tools.commands.xml_validate.sparql.resource_shacl import construct_resource_class_node_shape


def construct_shapes_graph(onto: Graph) -> Graph:
    """
    Constructs a shapes graph from a project ontology

    Args:
        onto: ontology as graph

    Returns:
        shapes graph
    """
    return construct_resource_class_node_shape(onto)



from __future__ import annotations

from datetime import datetime
from typing import Any
from typing import cast

from loguru import logger

from dsp_tools.commands.xmlupload.models.upload_state import UploadState
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStash
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStashItem
from dsp_tools.commands.xmlupload.stash.stash_models import Stash
from dsp_tools.models.exceptions import BaseError
from dsp_tools.utils.connection import Connection


def upload_stashed_resptr_props(
    upload_state: UploadState,
    con: Connection,
    context: dict[str, str],
) -> None:
    """
    After all resources are uploaded, the stashed resptr props must be applied to their resources in DSP.
    The upload state is updated accordingly, as a side effect.

    Args:
        upload_state: the current state of the upload
        con: connection to DSP
        context: the JSON-LD context of the resource
    """

    print(f"{datetime.now()}: Upload the stashed resptrs...")
    logger.info("Upload the stashed resptrs...")
    upload_state.pending_stash = cast(Stash, upload_state.pending_stash)
    link_value_stash = cast(LinkValueStash, upload_state.pending_stash.link_value_stash)
    for res_id, stash_items in link_value_stash.res_2_stash_items.copy().items():
        res_iri = upload_state.iri_resolver.get(res_id)
        if not res_iri:
            # resource could not be uploaded to DSP, so the stash cannot be uploaded either
            # no action necessary: this resource will remain in nonapplied_resptr_props,
            # which will be handled by the caller
            continue
        print(f"{datetime.now()}:   Upload resptrs of resource '{res_id}'...")
        logger.info(f"  Upload resptrs of resource '{res_id}'...")
        for stash_item in reversed(stash_items):
            # reversed avoids any problems caused by removing from the list we loop over at the same time
            target_iri = upload_state.iri_resolver.get(stash_item.target_id)
            if not target_iri:
                continue
            if _upload_stash_item(stash_item, res_iri, target_iri, con, context):
                link_value_stash.res_2_stash_items[res_id].remove(stash_item)
        # remove res_id if all stash items were uploaded
        if not link_value_stash.res_2_stash_items[res_id]:
            del link_value_stash.res_2_stash_items[res_id]


def _upload_stash_item(
    stash: LinkValueStashItem,
    res_iri: str,
    target_iri: str,
    con: Connection,
    context: dict[str, str],
) -> bool:
    """
    Upload a single stashed link value to DSP.

    Args:
        stash: the stashed link value to upload
        res_iri: the iri of the resource
        target_iri: the iri of the target resource
        con: connection to DSP
        context: the JSON-LD context of the resource

    Returns:
        True, if the upload was successful, False otherwise
    """
    payload = _create_resptr_prop_json_object_to_update(stash, res_iri, target_iri, context)
    try:
        con.post(route="/v2/values", data=payload)
    except BaseError as err:
        _log_unable_to_upload_link_value(err.message, stash.res_id, stash.prop_name)
        return False
    logger.debug(f'  Successfully uploaded resptr links of "{stash.prop_name}"')
    return True


def _log_unable_to_upload_link_value(msg: str, res_id: str, prop_name: str) -> None:
    err_msg = f"Unable to upload the resptr prop of '{prop_name}' of resource '{res_id}'."
    print(f"{datetime.now()}:     WARNING: {err_msg} Original error message: {msg}")
    logger.opt(exception=True).warning(err_msg)


def _create_resptr_prop_json_object_to_update(
    stash: LinkValueStashItem,
    res_iri: str,
    target_iri: str,
    context: dict[str, str],
) -> dict[str, Any]:
    """This function creates a JSON object that can be sent as an update request to the DSP-API."""
    linkVal = {
        "@type": "knora-api:LinkValue",
        "knora-api:linkValueHasTargetIri": {"@id": target_iri},
    }
    if stash.permission:
        linkVal["knora-api:hasPermissions"] = stash.permission
    jsonobj = {
        "@id": res_iri,
        "@type": stash.res_type,
        f"{stash.prop_name}Value": linkVal,
        "@context": context,
    }
    return jsonobj



from __future__ import annotations

import warnings
from collections import defaultdict
from typing import Any
from typing import cast

import pandas as pd
import regex
from loguru import logger

from dsp_tools.commands.excel2json.models.input_error import PositionInExcel
from dsp_tools.commands.excel2json.models.input_error import Problem
from dsp_tools.commands.excel2json.new_lists.models.deserialise import Columns
from dsp_tools.commands.excel2json.new_lists.models.deserialise import ExcelSheet
from dsp_tools.commands.excel2json.new_lists.models.input_error import CollectedSheetProblems
from dsp_tools.commands.excel2json.new_lists.models.input_error import DuplicateIDProblem
from dsp_tools.commands.excel2json.new_lists.models.input_error import DuplicatesCustomIDInProblem
from dsp_tools.commands.excel2json.new_lists.models.input_error import DuplicatesInSheetProblem
from dsp_tools.commands.excel2json.new_lists.models.input_error import DuplicatesListNameProblem
from dsp_tools.commands.excel2json.new_lists.models.input_error import ListCreationProblem
from dsp_tools.commands.excel2json.new_lists.models.input_error import ListInformation
from dsp_tools.commands.excel2json.new_lists.models.input_error import ListSheetComplianceProblem
from dsp_tools.commands.excel2json.new_lists.models.input_error import ListSheetContentProblem
from dsp_tools.commands.excel2json.new_lists.models.input_error import MinimumRowsProblem
from dsp_tools.commands.excel2json.new_lists.models.input_error import MissingExpectedColumn
from dsp_tools.commands.excel2json.new_lists.models.input_error import MissingNodeColumn
from dsp_tools.commands.excel2json.new_lists.models.input_error import MissingNodeTranslationProblem
from dsp_tools.commands.excel2json.new_lists.models.input_error import MissingTranslationsSheetProblem
from dsp_tools.commands.excel2json.new_lists.models.input_error import MultipleListPerSheetProblem
from dsp_tools.commands.excel2json.new_lists.models.input_error import NodesPerRowProblem
from dsp_tools.commands.excel2json.new_lists.models.input_error import SheetProblem
from dsp_tools.commands.excel2json.new_lists.utils import get_columns_of_preferred_lang
from dsp_tools.commands.excel2json.new_lists.utils import get_hierarchy_nums
from dsp_tools.commands.excel2json.new_lists.utils import get_lang_string_from_column_name
from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.models.exceptions import InputError


def make_all_excel_compliance_checks(sheet_list: list[ExcelSheet]) -> None:
    """Check if the excel files are compliant with the expected format."""
    # These functions must be called in this order,
    # as some of the following checks only work if the previous have passed.
    _check_duplicates_all_excels(sheet_list)
    _make_shape_compliance_all_excels(sheet_list)
    _check_for_missing_translations_all_excels(sheet_list)
    _check_for_unique_list_names(sheet_list)
    _check_for_erroneous_entries_all_excels(sheet_list)


def _check_duplicates_all_excels(sheet_list: list[ExcelSheet]) -> None:
    """
    Check if the excel files contain duplicates with regard to the node names,
    and if the custom IDs are unique across all excel files.
    A duplicate in the node names is defined as several rows with the same entries in the columns with the node names.

    Args:
        sheet_list: class instances representing an excel file with sheets

    Raises:
        InputError: If any complete duplicates are found in the excel files.
    """
    problems: list[Problem] = []
    duplicate_problems: list[SheetProblem] = [
        p for sheet in sheet_list if (p := _check_for_duplicate_nodes_one_df(sheet)) is not None
    ]
    if duplicate_problems:
        problems.append(CollectedSheetProblems(duplicate_problems))
    if id_problem := _check_for_duplicate_custom_id_all_excels(sheet_list):
        problems.append(id_problem)
    if problems:
        msg = ListCreationProblem(problems).execute_error_protocol()
        logger.error(msg)
        raise InputError(msg)


def _check_for_unique_list_names(sheet_list: list[ExcelSheet]) -> None:
    """This functon checks that one sheet only has one list and that all lists have unique names."""
    list_names: list[ListInformation] = []
    all_problems: list[Problem] = []
    sheet_problems: list[SheetProblem] = []
    for sheet in sheet_list:
        unique_list_names = list(sheet.df[f"{sheet.col_info.preferred_lang}_list"].unique())
        if len(unique_list_names) != 1:
            sheet_problems.append(MultipleListPerSheetProblem(sheet.excel_name, sheet.sheet_name, unique_list_names))
        list_names.extend([ListInformation(sheet.excel_name, sheet.sheet_name, name) for name in unique_list_names])
    if sheet_problems:
        all_problems.append(CollectedSheetProblems(sheet_problems))
    list_info_dict = defaultdict(list)
    for item in list_names:
        list_info_dict[item.list_name].append(item)
    duplicate_list_names = []
    for info in list_info_dict.values():
        if len(info) > 1:
            duplicate_list_names.extend(info)
    if duplicate_list_names:
        all_problems.append(DuplicatesListNameProblem(duplicate_list_names))
    if all_problems:
        msg = ListCreationProblem(all_problems).execute_error_protocol()
        logger.error(msg)
        raise InputError(msg)


def _check_for_duplicate_nodes_one_df(sheet: ExcelSheet) -> DuplicatesInSheetProblem | None:
    """Check if any rows have duplicates when taking into account the columns with the node names."""
    lang_columns = [col for col in sheet.df.columns if regex.search(r"^(en|de|fr|it|rm)_(\d+|list)$", col)]
    if (duplicate_filter := sheet.df.duplicated(lang_columns, keep=False)).any():
        return DuplicatesInSheetProblem(
            sheet.excel_name, sheet.sheet_name, duplicate_filter.index[duplicate_filter].tolist()
        )
    return None


def _check_for_duplicate_custom_id_all_excels(sheet_list: list[ExcelSheet]) -> DuplicatesCustomIDInProblem | None:
    id_list = []
    for sheet in sheet_list:
        for i, row in sheet.df.iterrows():
            if not pd.isna(row["id (optional)"]):
                id_list.append(
                    {
                        "filename": sheet.excel_name,
                        "sheet_name": sheet.sheet_name,
                        "id": row["id (optional)"],
                        "row_number": int(str(i)) + 2,
                    }
                )
    id_df = pd.DataFrame.from_records(id_list)
    if (duplicate_ids := id_df.duplicated("id", keep=False)).any():
        problems: dict[str, DuplicateIDProblem] = defaultdict(lambda: DuplicateIDProblem())
        for i, row in id_df[duplicate_ids].iterrows():
            problems[row["id"]].custom_id = row["id"]
            problems[row["id"]].excel_locations.append(
                PositionInExcel(sheet=row["sheet_name"], excel_filename=row["filename"], row=row["row_number"])
            )
        final_problems = list(problems.values())
        return DuplicatesCustomIDInProblem(final_problems)
    return None


def _make_shape_compliance_all_excels(sheet_list: list[ExcelSheet]) -> None:
    """Check if the excel files are compliant with the expected format."""
    problems: list[SheetProblem] = [
        p for sheet in sheet_list if (p := _make_shape_compliance_one_sheet(sheet)) is not None
    ]
    if problems:
        msg = ListCreationProblem([CollectedSheetProblems(problems)]).execute_error_protocol()
        logger.error(msg)
        raise InputError(msg)


def _make_shape_compliance_one_sheet(sheet: ExcelSheet) -> ListSheetComplianceProblem | None:
    problems: list[Problem] = []
    if len(sheet.df) < 2:
        problems.append(MinimumRowsProblem())
    if not sheet.col_info.node_cols:
        problems.append(MissingNodeColumn())
    if missing := _check_if_all_translations_in_all_column_levels_present_one_sheet(sheet.df.columns):
        problems.append(missing)
    _check_warn_unusual_columns_one_sheet(sheet.df.columns)
    if problems:
        return ListSheetComplianceProblem(sheet.excel_name, sheet.sheet_name, problems)
    return None


def _check_warn_unusual_columns_one_sheet(cols: pd.Index[str]) -> None:
    not_matched = [x for x in cols if not regex.search(r"^(en|de|fr|it|rm)_(\d+|list|comments)|(id \(optional\))$", x)]
    if not_matched:
        msg = (
            f"The following columns do not conform to the expected format "
            f"and will not be included in the output: {', '.join(not_matched)}"
        )
        warnings.warn(DspToolsUserWarning(msg))


def _check_if_all_translations_in_all_column_levels_present_one_sheet(
    cols: pd.Index[str],
) -> MissingExpectedColumn | None:
    # All levels, eg. 1, 2, 3 must have the same languages
    languages = {r for c in cols if (r := get_lang_string_from_column_name(c)) is not None}
    all_nums = [str(n) for n in get_hierarchy_nums(cols)]
    all_nums.append("list")

    def make_col_names(lang: str) -> set[str]:
        return {f"{lang}_{num}" for num in all_nums}

    expected_cols = set()
    for lang in languages:
        expected_cols.update(make_col_names(lang))
    if missing_cols := expected_cols - set(cols):
        return MissingExpectedColumn(missing_cols)
    return None


def _check_for_missing_translations_all_excels(sheet_list: list[ExcelSheet]) -> None:
    problems: list[SheetProblem] = [
        p for sheet in sheet_list if (p := _check_for_missing_translations_one_sheet(sheet)) is not None
    ]
    if problems:
        msg = ListCreationProblem([CollectedSheetProblems(problems)]).execute_error_protocol()
        logger.error(msg)
        raise InputError(msg)


def _check_for_missing_translations_one_sheet(sheet: ExcelSheet) -> MissingTranslationsSheetProblem | None:
    problems = []
    for i, row in sheet.df.iterrows():
        if problem := _check_missing_translations_one_row(int(str(i)), row, sheet.col_info):
            problems.append(problem)
    if problems:
        return MissingTranslationsSheetProblem(sheet.excel_name, sheet.sheet_name, problems)
    return None


def _check_missing_translations_one_row(
    row_index: int, row: pd.Series[Any], columns: Columns
) -> MissingNodeTranslationProblem | None:
    missing_translations = []
    for col_group in columns.node_cols:
        missing_translations.extend(_check_for_missing_translations_one_column_group(row, col_group.columns))
    missing_translations.extend(_check_for_missing_translations_one_column_group(row, columns.list_cols))
    if columns.comment_cols:
        missing_translations.extend(_check_for_missing_translations_one_column_group(row, columns.comment_cols))
    if missing_translations:
        return MissingNodeTranslationProblem(empty_columns=missing_translations, index_num=row_index)
    return None


def _check_for_missing_translations_one_column_group(row: pd.Series[Any], columns: list[str]) -> list[str]:
    missing = row[columns].isna()
    if missing.any() and not missing.all():
        return [str(index) for index, is_missing in missing.items() if is_missing]
    return []


def _check_for_erroneous_entries_all_excels(sheet_list: list[ExcelSheet]) -> None:
    problems: list[SheetProblem] = [
        p for sheet in sheet_list if (p := _check_for_erroneous_entries_one_list(sheet)) is not None
    ]
    if problems:
        msg = ListCreationProblem([CollectedSheetProblems(problems)]).execute_error_protocol()
        logger.error(msg)
        raise InputError(msg)


def _check_for_erroneous_entries_one_list(sheet: ExcelSheet) -> ListSheetContentProblem | None:
    preferred_cols = get_columns_of_preferred_lang(sheet.df.columns, sheet.col_info.preferred_lang, r"\d+")
    preferred_cols = sorted(preferred_cols)
    preferred_cols.insert(0, f"{sheet.col_info.preferred_lang}_list")
    problems = _check_for_erroneous_node_info_one_df(sheet.df, preferred_cols)
    if problems:
        list_problems = cast(list[Problem], problems)
        return ListSheetContentProblem(sheet.excel_name, sheet.sheet_name, list_problems)
    return None


def _check_for_erroneous_node_info_one_df(df: pd.DataFrame, columns: list[str]) -> list[NodesPerRowProblem]:
    problems = []
    for focus_col_index, col in enumerate(columns):
        problems.extend(_check_for_erroneous_entries_one_column_level(df, columns, focus_col_index))
    return problems


def _check_for_erroneous_entries_one_column_level(
    df: pd.DataFrame, columns: list[str], focus_col_index: int
) -> list[NodesPerRowProblem]:
    # column level refers to the hierarchical level of the nodes. eg. "en_1"
    # we need to group by from the current column all the way back to its ancestors,
    # otherwise identical values in that column may be interpreted as belonging to the same group
    grouped = df.groupby(columns[: focus_col_index + 1])
    problems = []
    for name, group in grouped:
        remaining_to_check_columns = columns[focus_col_index:]
        problems.extend(_check_for_erroneous_entries_one_grouped_df(group, remaining_to_check_columns))
    return problems


def _check_for_erroneous_entries_one_grouped_df(
    group: pd.DataFrame, target_cols: list[str]
) -> list[NodesPerRowProblem]:
    problems = []
    first_col = min(group.index)
    # The first row is the current parent node. The remaining columns in that row must be empty.
    if not group.loc[first_col, target_cols[1:]].isna().all():
        problems.append(NodesPerRowProblem(target_cols[1:], int(first_col), should_be_empty=True))
    if not len(target_cols) > 1:
        return problems
    # The second column of the remaining rows must not be empty, as these are the child nodes of the first row.
    remaining_rows_of_next_column = group.loc[group.index[1:], target_cols[1]]
    if (missing := remaining_rows_of_next_column.isna()).any():
        for i, row in group[1:][missing].iterrows():
            problems.append(NodesPerRowProblem([target_cols[1]], int(str(i)), should_be_empty=False))
    return problems



"""end to end tests for listnode class"""

import unittest

import pytest

from dsp_tools.commands.project.models.listnode import ListNode
from dsp_tools.models.langstring import LangString
from dsp_tools.models.langstring import Languages
from dsp_tools.utils.connection import Connection
from dsp_tools.utils.connection_live import ConnectionLive

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)


class TestListNode(unittest.TestCase):
    con: Connection
    project = "http://rdfh.ch/projects/0001"
    otherTreeList = "http://rdfh.ch/lists/0001/otherTreeList"

    def setUp(self) -> None:
        """
        Creates a connection to DSP-API.
        For each test method, a new TestCase instance is created, so setUp() is executed before each test method.
        """
        self.con = ConnectionLive(server="http://0.0.0.0:3333")
        self.con.login(email="root@example.com", password="test")

    def tearDown(self) -> None:
        """
        Logs out from DSP-API.
        For each test method, a new TestCase instance is created, so tearDown() is executed after each test method.
        """
        self.con.logout()

    def test_ListNode_read(self) -> None:
        """
        Read an existing node
        :return: None
        """
        node = ListNode(con=self.con, iri=self.otherTreeList).read()
        self.assertEqual(node.iri, self.otherTreeList)
        self.assertEqual(node.project, self.project)
        self.assertEqual(node.label["en"], "Tree list root")
        self.assertTrue(node.isRootNode)
        self.assertEqual(node.children, [])

    def test_ListNode_create(self) -> None:
        """
        Create a list node
        :return: None
        """
        node = ListNode(
            con=self.con,
            project=self.project,
            label=LangString({Languages.DE: "root node 1"}),
            comments=LangString({Languages.DE: "first root node"}),
            name="test_node_1",
        ).create()
        self.assertIsNotNone(node.iri)
        self.assertEqual(node.project, self.project)
        self.assertEqual(node.label["de"], "root node 1")
        self.assertEqual(node.comments["de"], "first root node")
        self.assertEqual(node.name, "test_node_1")
        self.assertTrue(node.isRootNode)

    def test_ListNode_hierarchy(self) -> None:
        """
        Create a node and a sub-node
        :return: None
        """
        node = ListNode(
            con=self.con,
            project=self.project,
            label=LangString({Languages.EN: "root node"}),
            comments=LangString({Languages.EN: "This is a root node"}),
            name="root_node",
        ).create()
        subnode = ListNode(
            con=self.con,
            project=self.project,
            label=LangString({Languages.DE: "Ein Knoten der Liste"}),
            comments=LangString({Languages.DE: "Ein Kommentar"}),
            name="sub_node",
            parent=node,
        ).create()
        self.assertTrue(node.isRootNode)
        self.assertFalse(subnode.isRootNode)

    def test_ListNode_getAllLists(self) -> None:
        """
        Get all lists
        :return: None
        """
        lists = ListNode.getAllLists(self.con, self.project)
        list_ids = list(map(lambda lst: lst.iri, lists))
        self.assertIn(self.otherTreeList, list_ids)
        self.assertIn("http://rdfh.ch/lists/0001/treeList", list_ids)

    def test_ListNode_getAllNodes(self) -> None:
        """
        Get all nodes of a list
        :return: None
        """
        root_node = ListNode(con=self.con, iri=self.otherTreeList).getAllNodes()

        self.assertTrue(root_node.isRootNode)
        self.assertEqual(root_node.project, self.project)
        self.assertEqual(root_node.label["en"], "Tree list root")
        self.assertIsNotNone(root_node.children)

        self.assertEqual(root_node.children[0].iri, "http://rdfh.ch/lists/0001/otherTreeList01")
        self.assertEqual(root_node.children[0].name, "Other Tree list node 01")
        self.assertEqual(root_node.children[0].label["en"], "Other Tree list node 01")

        self.assertEqual(root_node.children[1].iri, "http://rdfh.ch/lists/0001/otherTreeList02")
        self.assertEqual(root_node.children[1].name, "Other Tree list node 02")
        self.assertEqual(root_node.children[1].label["en"], "Other Tree list node 02")

        self.assertEqual(root_node.children[2].iri, "http://rdfh.ch/lists/0001/otherTreeList03")
        self.assertEqual(root_node.children[2].name, "Other Tree list node 03")
        self.assertEqual(root_node.children[2].label["en"], "Other Tree list node 03")

        self.assertIsNotNone(root_node.children[2].children)
        self.assertEqual(root_node.children[2].children[0].iri, "http://rdfh.ch/lists/0001/otherTreeList10")
        self.assertEqual(root_node.children[2].children[0].name, "Other Tree list node 10")
        self.assertEqual(root_node.children[2].children[0].label["en"], "Other Tree list node 10")

        self.assertEqual(root_node.children[2].children[1].iri, "http://rdfh.ch/lists/0001/otherTreeList11")
        self.assertEqual(root_node.children[2].children[1].name, "Other Tree list node 11")
        self.assertEqual(root_node.children[2].children[1].label["en"], "Other Tree list node 11")


if __name__ == "__main__":
    pytest.main([__file__])



from rdflib.namespace import Namespace


def get_default_json_ld_context() -> dict[str, str]:
    """
    Returns the JSON-LD context as a dictionary.

    Returns:
        JSON-LD context as a dictionary.
    """
    return {
        "knora-api": "http://api.knora.org/ontology/knora-api/v2#",
        "salsah-gui": "http://api.knora.org/ontology/salsah-gui/v2#",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "owl": "http://www.w3.org/2002/07/owl#",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
    }


def make_namespace_dict_from_onto_names(ontos: dict[str, str]) -> dict[str, Namespace]:
    """Provided a dictionary of ontology names and IRIs, returns a dictionary of Namespace objects."""
    ontos = correct_project_context_namespaces(ontos)
    namespaces = {k: Namespace(v) for k, v in ontos.items()}
    namespaces["knora-api"] = Namespace("http://api.knora.org/ontology/knora-api/v2#")
    return namespaces


def correct_project_context_namespaces(ontos: dict[str, str]) -> dict[str, str]:
    """Add the hashtag to make it a valid namespace."""
    return {k: f"{v}#" for k, v in ontos.items()}


def get_json_ld_context_for_project(ontos: dict[str, str]) -> dict[str, str]:
    """Provided a dictionary of ontology names and IRIs, returns a JSON-LD context for the project."""
    context = get_default_json_ld_context()
    project_context = correct_project_context_namespaces(ontos)
    context.update(project_context)
    return context



"""This module handles all the operations which are used for the creation of JSON lists from Excel files."""

import importlib.resources
import json
from copy import deepcopy
from pathlib import Path
from typing import Any
from typing import Optional
from typing import Union
from typing import cast

import jsonschema
import regex
from openpyxl import load_workbook
from openpyxl.cell import Cell
from openpyxl.worksheet.worksheet import Worksheet

from dsp_tools.commands.excel2json.models.input_error import MoreThanOneSheetProblem
from dsp_tools.commands.excel2json.models.list_node_name import ListNodeNames
from dsp_tools.models.exceptions import BaseError
from dsp_tools.models.exceptions import InputError
from dsp_tools.models.exceptions import UserError
from dsp_tools.utils.shared import simplify_name


def expand_lists_from_excel(
    lists_section: list[dict[str, Union[str, dict[str, Any]]]],
) -> list[dict[str, Any]]:
    """
    Checks if the "lists" section of a JSON project file contains references to Excel files.
    Expands all Excel files to JSON,
    and returns the expanded "lists" section.
    If there are no references to Excel files,
    the "lists" section is returned as is.

    Args:
        lists_section: the "lists" section of a parsed JSON project file.
            If this is an empty list, an empty list will be returned.

    Raises:
        UserError: if a problem occurred while trying to expand the Excel files

    Returns:
        the same "lists" section, but without references to Excel files
    """
    new_lists: list[dict[str, Any]] = []
    for _list in lists_section:
        # case 1: this list is a JSON list: return it as it is
        if "folder" not in _list["nodes"]:
            new_lists.append(_list)
            continue

        # case 2: this is a reference to a folder with Excel files
        new_lists.append(_read_excel_make_lists(_list))

    return new_lists


def _read_excel_make_lists(_list: dict[str, Union[str, dict[str, Any]]]) -> dict[str, Any]:
    foldername = _list["nodes"]["folder"]  # type: ignore[index]  # types are too complex to annotate them correctly
    excel_file_paths = _extract_excel_file_paths(foldername)
    new_list = deepcopy(_list)
    try:
        returned_lists_section = _make_json_lists_from_excel(excel_file_paths, verbose=False)
        # we only need the "nodes" section of the first element of the returned "lists" section. This "nodes"
        # section needs to replace the Excel folder reference. But the rest of the user-defined list element
        # needs to stay intact, e.g. the labels and comments.
        new_list["nodes"] = returned_lists_section[0]["nodes"]
        print(
            f"    The list '{_list['name']}' contains a reference to the folder '{foldername}'. The Excel "
            f"files therein have been temporarily expanded into the 'lists' section of your project."
        )
        return new_list
    except BaseError as err:
        raise UserError(
            f"    WARNING: The list '{_list['name']}' contains a reference to the folder '{foldername}', but a "
            f"problem occurred while trying to expand the Excel files therein into the 'lists' section of "
            f"your project: {err.message}"
        ) from None


def excel2lists(
    excelfolder: str,
    path_to_output_file: Optional[str] = None,
    verbose: bool = False,
) -> tuple[list[dict[str, Any]], bool]:
    """
    Converts lists described in Excel files into a "lists" section that can be inserted into a JSON project file.

    Args:
        excelfolder: path to the folder containing the Excel file(s)
        path_to_output_file: if provided, the output is written into this JSON file
        verbose: verbose switch

    Raises:
        UserError: if something went wrong
        BaseError: if something went wrong

    Returns:
        a tuple consisting of the "lists" section as Python list, and the success status (True if everything went well)
    """
    excel_file_paths = _extract_excel_file_paths(excelfolder)
    if verbose:
        print("The following Excel files will be processed:")
        print(*(f" - {filename}" for filename in excel_file_paths), sep="\n")

    finished_lists = _make_json_lists_from_excel(excel_file_paths, verbose=verbose)
    validate_lists_section_with_schema(lists_section=finished_lists)

    if path_to_output_file:
        with open(path_to_output_file, "w", encoding="utf-8") as fp:
            json.dump(finished_lists, fp, indent=4, ensure_ascii=False)
            print(f"lists section was created successfully and written to file '{path_to_output_file}'")

    return finished_lists, True


def _get_values_from_excel(
    excelfiles: dict[str, Worksheet],
    base_file: dict[str, Worksheet],
    parentnode: dict[str, Any],
    row: int,
    col: int,
    preval: list[str],
    list_node_names: ListNodeNames,
    verbose: bool = False,
) -> tuple[int, dict[str, Any]]:
    """
    This function calls itself recursively to go through the Excel files. It extracts the cell values and composes
    the JSON list.

    Args:
        excelfiles: List of Excel files with the values in different languages
        base_file: File name of the base file
        parentnode: Name(s) of the parent node(s) of the current node
        row: The index of the current row of the Excel sheet
        col: The index of the current column of the Excel sheet
        preval: List of previous values, needed to check the consistency of the list hierarchy
        list_node_names: object containing the information which node names were used by previous lists
        verbose: verbose switch

    Raises:
        UserError: if one of the Excel files contains invalid data

    Returns:
        int: Row index for the next loop (current row index minus 1)
        dict: The JSON list up to the current recursion. At the last recursion, this is the final JSON list.
    """
    nodes: list[dict[str, Any]] = []
    currentnode: dict[str, Any] = {}
    base_file_ws: Worksheet = next(iter(base_file.values()))
    cell: Cell = base_file_ws.cell(column=col, row=row)

    for excelfile in excelfiles.values():
        if any((not excelfile["A1"].value, excelfile["B1"].value)):
            raise UserError(
                f"ERROR: Inconsistency in Excel list: The first row must consist of exactly one value, in cell A1. "
                f"All other cells of row 1 must be empty.\nInstead, found the following:\n"
                f" - Cell A1: '{excelfile['A1'].value}'\n"
                f" - Cell B1: '{excelfile['B1'].value}'"
            )

    if col > 1:
        # append the cell value of the parent node (which is one value to the left of the current cell) to the list of
        # previous values
        preval.append(str(base_file_ws.cell(column=col - 1, row=row).value).strip())

    while cell.value and regex.search(r"\p{L}", str(cell.value), flags=regex.UNICODE):
        _check_if_predecessors_in_row_are_consistent_with_preval(base_file_ws, preval, row)

        # loop through the row until the last (furthest right) value is found
        next_value = base_file_ws.cell(column=col + 1, row=row).value
        if next_value and regex.search(r"\p{L}", str(next_value), flags=regex.UNICODE):
            row, _ = _get_values_from_excel(
                excelfiles=excelfiles,
                base_file=base_file,
                parentnode=currentnode,
                col=col + 1,
                row=row,
                preval=preval,
                list_node_names=list_node_names,
                verbose=verbose,
            )

        # if value was last in row (no further values to the right), it's a node, continue here
        else:
            currentnode = _make_new_node(cell, col, excelfiles, preval, row, list_node_names, verbose)
            nodes.append(currentnode)

        # go one row down and repeat loop if there is a value
        row += 1
        cell = base_file_ws.cell(column=col, row=row)

    if col > 1:
        preval.pop()

    # add the new nodes to the parentnode
    parentnode["nodes"] = nodes

    return row - 1, parentnode


def _check_if_predecessors_in_row_are_consistent_with_preval(
    base_file_ws: Worksheet, preval: list[str], row: int
) -> None:
    for idx, val in enumerate(preval[:-1]):
        if val != str(base_file_ws.cell(column=idx + 1, row=row).value).strip():
            raise UserError(
                "ERROR: Inconsistency in Excel list: "
                f"{val} not equal to {str(base_file_ws.cell(column=idx + 1, row=row).value).strip()}"
            )


def _make_new_node(
    cell: Cell,
    col: int,
    excelfiles: dict[str, Worksheet],
    preval: list[str],
    row: int,
    list_node_names: ListNodeNames,
    verbose: bool = False,
) -> dict[str, Any]:
    _check_if_duplicate_nodes_exist(cell, list_node_names, preval)
    # create a simplified version of the cell value and use it as name of the node
    nodename = simplify_name(str(cell.value).strip())
    list_node_names.previous_node_names.append(nodename)
    # append a number (p.ex. node-name-2) if there are list nodes with identical names
    n = list_node_names.previous_node_names.count(nodename)

    if n > 1:
        nodename = f"{nodename}-{n}"

    labels_dict = _get_all_languages_of_node(excelfiles, col, row)

    # create the current node from extracted cell values and append it to the nodes list
    currentnode = {"name": nodename, "labels": labels_dict}
    if verbose:
        print(f"Added list node: {str(cell.value).strip()} ({nodename})")
    return currentnode


def _check_if_duplicate_nodes_exist(cell: Cell, list_node_names: ListNodeNames, preval: list[str]) -> None:
    new_check_list = preval.copy()
    new_check_list.append(str(cell.value).strip())
    list_node_names.lists_with_previous_cell_values.append(new_check_list)
    if any(
        list_node_names.lists_with_previous_cell_values.count(x) > 1
        for x in list_node_names.lists_with_previous_cell_values
    ):
        raise UserError(
            f"ERROR: There is at least one duplicate node in the list. "
            f"Found duplicate in column {cell.column}, row {cell.row}:\n'{str(cell.value).strip()}'"
        )


def _get_all_languages_of_node(excelfiles: dict[str, Worksheet], col: int, row: int) -> dict[str, str]:
    labels_dict: dict[str, str] = {}
    for other_lang, ws_other_lang in excelfiles.items():
        cell_value = ws_other_lang.cell(column=col, row=row).value
        if not (isinstance(cell_value, str) and len(cell_value) > 0):
            raise UserError(
                "ERROR: Malformed Excel file: The Excel file with the language code "
                f"'{other_lang}' should have a value in row {row}, column {col}"
            )
        else:
            labels_dict[other_lang] = cell_value.strip()
    return labels_dict


def _make_json_lists_from_excel(
    excel_file_paths: list[Path],
    verbose: bool = False,
) -> list[dict[str, Any]]:
    """
    Reads Excel files and transforms them into a list of dictionaries that can be used as "lists" array of a JSON
    project file.

    Args:
        excel_file_paths: Excel files to be processed
        verbose: verbose switch

    Raises:
        UserError: if one of the Excel files contains invalid data

    Returns:
        The finished "lists" section
    """

    # Define starting point in Excel file
    startrow = 1
    startcol = 1

    lang_to_worksheet: dict[str, Worksheet] = {x.stem: _read_and_check_workbook(x) for x in excel_file_paths}

    base_lang = "en" if "en" in lang_to_worksheet else next(iter(lang_to_worksheet.keys()))
    base_file = {base_lang: lang_to_worksheet[base_lang]}

    list_node_names = ListNodeNames()

    # construct the entire "lists" section as children of a fictive dummy parent node
    _, _list = _get_values_from_excel(
        excelfiles=lang_to_worksheet,
        base_file=base_file,
        parentnode={},
        row=startrow,
        col=startcol,
        preval=[],
        list_node_names=list_node_names,
        verbose=verbose,
    )

    # extract the children of the fictive dummy parent node
    finished_lists: list[dict[str, Any]] = _list["nodes"]

    # the "comments" section is mandatory for the root node of each list, so make a copy of "labels" and insert it at
    # the right place
    for i, _list in enumerate(finished_lists):
        finished_lists[i] = {
            "name": _list["name"],
            "labels": _list["labels"],
            "comments": _list["labels"],
            "nodes": _list["nodes"],
        }

    return finished_lists


def _read_and_check_workbook(excelpath: Path) -> Worksheet:
    all_worksheets = cast(list[Worksheet], load_workbook(excelpath, read_only=True).worksheets)
    if len(all_worksheets) != 1:
        msg = MoreThanOneSheetProblem(excelpath.name, [x.title for x in all_worksheets]).execute_error_protocol()
        raise InputError(msg)
    return all_worksheets[0]


def validate_lists_section_with_schema(
    path_to_json_project_file: Optional[str] = None,
    lists_section: Optional[list[dict[str, Any]]] = None,
) -> bool:
    """
    This function checks if a "lists" section of a JSON project is valid according to the schema. The "lists" section
    can be passed as path to the JSON project file, or as Python object. Only one of the two arguments should be passed.

    Args:
        path_to_json_project_file: path to the JSON project file that contains the "lists" section to validate
        lists_section: the "lists" section as Python object

    Raises:
        UserError: if the validation fails
        BaseError: if this function is called with invalid parameters

    Returns:
        True if the "lists" section passed validation
    """
    err_msg = "Validation of the 'lists' section works only if exactly one of the two arguments is given."
    match path_to_json_project_file, lists_section:
        case None, None:
            raise BaseError(err_msg)
        case str(), list():
            raise BaseError(err_msg)

    with (
        importlib.resources.files("dsp_tools")
        .joinpath("resources/schema/lists-only.json")
        .open(encoding="utf-8") as schema_file
    ):
        lists_schema = json.load(schema_file)

    if path_to_json_project_file:
        with open(path_to_json_project_file, encoding="utf-8") as f:
            project = json.load(f)
            lists_section = project["project"].get("lists")
            if not lists_section:
                raise UserError(
                    f"Cannot validate 'lists' section of {path_to_json_project_file}, "
                    "because there is no 'lists' section in this file."
                )

    try:
        jsonschema.validate(instance={"lists": lists_section}, schema=lists_schema)
    except jsonschema.ValidationError as err:
        raise UserError(
            f"'lists' section did not pass validation. The error message is: {err.message}\n"
            f"The error occurred at {err.json_path}"
        ) from None

    return True


def _extract_excel_file_paths(excelfolder: str) -> list[Path]:
    """
    This method extracts the names of the Excel files that are in the folder, and asserts that they are named according
    to the requirements.

    Args:
        excelfolder: path to the folder containing the Excel file(s)

    Raises:
        UserError: if excelfolder is not a directory, or if one of the files in it has an invalid name

    Returns:
        list of the Excel file paths to process
    """
    if not Path(excelfolder).is_dir():
        raise UserError(f"ERROR: {excelfolder} is not a directory.")

    supported_files = ["en.xlsx", "de.xlsx", "fr.xlsx", "it.xlsx", "rm.xlsx"]
    excel_file_paths = [x for x in Path(excelfolder).glob("*.xlsx") if x.is_file() and not x.name.startswith("~$")]

    for filepath in excel_file_paths:
        if filepath.name not in supported_files:
            raise UserError(f"Invalid file name '{filepath}'. Expected format: 'languagecode.xlsx'")

    return excel_file_paths



import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from typing import assert_never
from typing import cast

from loguru import logger
from rdflib import RDF
from rdflib import BNode
from rdflib import Graph
from rdflib import Literal
from rdflib import Namespace
from rdflib import URIRef

from dsp_tools.commands.xmlupload.ark2iri import convert_ark_v0_to_resource_iri
from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import IIIFUriInfo
from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLProperty
from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLValue
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import BitstreamInfo
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import XMLResource
from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.namespace_context import get_json_ld_context_for_project
from dsp_tools.commands.xmlupload.models.namespace_context import make_namespace_dict_from_onto_names
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.serialise.jsonld_serialiser import serialise_property_graph
from dsp_tools.commands.xmlupload.models.serialise.serialise_value import SerialiseProperty
from dsp_tools.commands.xmlupload.models.serialise.serialise_value import SerialiseURI
from dsp_tools.commands.xmlupload.models.serialise.serialise_value import SerialiseValue
from dsp_tools.models.exceptions import BaseError
from dsp_tools.models.exceptions import InputError
from dsp_tools.models.exceptions import PermissionNotExistsError
from dsp_tools.models.exceptions import UserError
from dsp_tools.utils.connection import Connection
from dsp_tools.utils.date_util import parse_date_string
from dsp_tools.utils.iri_util import is_resource_iri
from dsp_tools.utils.logger_config import WARNINGS_SAVEPATH

KNORA_API = Namespace("http://api.knora.org/ontology/knora-api/v2#")


@dataclass(frozen=True)
class ResourceCreateClient:
    """client class that creates resources on a DSP server."""

    con: Connection
    project_iri: str
    iri_resolver: IriResolver
    project_onto_dict: dict[str, str]
    permissions_lookup: dict[str, Permissions]
    listnode_lookup: dict[str, str]
    media_previously_ingested: bool = False

    def create_resource(
        self,
        resource: XMLResource,
        bitstream_information: BitstreamInfo | None,
    ) -> str:
        """Creates a resource on the DSP server, and returns its IRI"""
        logger.info(
            f"Attempting to create resource {resource.res_id} (label: {resource.label}, iri: {resource.iri})..."
        )
        resource_dict = self._make_resource_with_values(resource, bitstream_information)
        headers = {"X-Asset-Ingested": "true"} if bitstream_information else None
        res = self.con.post(route="/v2/resources", data=resource_dict, headers=headers)
        return cast(str, res["@id"])

    def _make_resource_with_values(
        self,
        resource: XMLResource,
        bitstream_information: BitstreamInfo | None,
    ) -> dict[str, Any]:
        res_bnode = BNode()
        namespaces = make_namespace_dict_from_onto_names(self.project_onto_dict)
        res = self._make_resource(
            resource=resource,
            bitstream_information=bitstream_information,
        )
        vals = self._make_values(resource, res_bnode, namespaces)
        res.update(vals)
        return res

    def _make_resource(
        self,
        resource: XMLResource,
        bitstream_information: BitstreamInfo | None,
    ) -> dict[str, Any]:
        resource_iri = resource.iri
        if resource.ark:
            resource_iri = convert_ark_v0_to_resource_iri(resource.ark)
        context = get_json_ld_context_for_project(self.project_onto_dict)
        res = {
            "@type": resource.restype,
            "rdfs:label": resource.label,
            "knora-api:attachedToProject": {"@id": self.project_iri},
            "@context": context,
        }
        if resource_iri:
            res["@id"] = resource_iri
        if resource.permissions:
            if perm := self.permissions_lookup.get(resource.permissions):
                res["knora-api:hasPermissions"] = str(perm)
            else:
                raise PermissionNotExistsError(
                    f"Could not find permissions for resource {resource.res_id} with permissions {resource.permissions}"
                )
        if resource.creation_date:
            res["knora-api:creationDate"] = {
                "@type": "xsd:dateTimeStamp",
                "@value": str(resource.creation_date),
            }
        if bitstream_information:
            res.update(_make_bitstream_file_value(bitstream_information))
        return res

    def _make_values(self, resource: XMLResource, res_bnode: BNode, namespaces: dict[str, Namespace]) -> dict[str, Any]:
        def prop_name(p: XMLProperty) -> str:
            if p.valtype != "resptr":
                return p.name
            elif p.name == "knora-api:isSegmentOf" and resource.restype == "knora-api:VideoSegment":
                return "knora-api:isVideoSegmentOfValue"
            elif p.name == "knora-api:isSegmentOf" and resource.restype == "knora-api:AudioSegment":
                return "knora-api:isAudioSegmentOfValue"
            else:
                return f"{p.name}Value"

        def make_values(p: XMLProperty) -> list[dict[str, Any]]:
            return [self._make_value(v, p.valtype) for v in p.values]

        properties_serialised = {}
        properties_graph = Graph()
        # To frame the json-ld correctly, we need one property used in the graph. It does not matter which.
        last_prop_name = None

        for prop in resource.properties:
            match prop.valtype:
                case "uri":
                    properties_serialised.update(_serialise_uri_prop(prop, self.permissions_lookup))
                case "integer":
                    int_prop_name = self._get_absolute_prop_iri(prop.name, namespaces)
                    int_graph = _make_integer_prop(prop, res_bnode, int_prop_name, self.permissions_lookup)
                    properties_graph += int_graph
                    last_prop_name = int_prop_name
                case "boolean":
                    bool_prop_name = self._get_absolute_prop_iri(prop.name, namespaces)
                    bool_graph = _make_boolean_prop(prop, res_bnode, bool_prop_name, self.permissions_lookup)
                    properties_graph += bool_graph
                    last_prop_name = bool_prop_name
                case _:
                    properties_serialised[prop_name(prop)] = make_values(prop)
        if resource.iiif_uri:
            properties_graph += _make_iiif_uri_value(resource.iiif_uri, res_bnode, self.permissions_lookup)
            last_prop_name = KNORA_API.hasStillImageFileValue
        if last_prop_name:
            serialised_graph_props = serialise_property_graph(properties_graph, last_prop_name)
            properties_serialised.update(serialised_graph_props)
        return properties_serialised

    def _get_absolute_prop_iri(self, prefixed_prop: str, namespaces: dict[str, Namespace]) -> URIRef:
        prefix, prop = prefixed_prop.split(":", maxsplit=1)
        if not (namespace := namespaces.get(prefix)):
            raise InputError(f"Could not find namespace for prefix: {prefix}")
        return namespace[prop]

    def _make_value(self, value: XMLValue, value_type: str) -> dict[str, Any]:  # noqa: PLR0912 (too many branches)
        match value_type:
            case "color":
                res = _make_color_value(value)
            case "date":
                res = _make_date_value(value)
            case "decimal":
                res = _make_decimal_value(value)
            case "geometry":
                res = _make_geometry_value(value)
            case "geoname":
                res = _make_geoname_value(value)
            case "interval":
                res = _make_interval_value(value)
            case "resptr":
                res = _make_link_value(value, self.iri_resolver)
            case "list":
                res = _make_list_value(value, self.listnode_lookup)
            case "text":
                res = _make_text_value(value, self.iri_resolver)
            case "time":
                res = _make_time_value(value)
            case _:
                raise UserError(f"Unknown value type: {value_type}")
        if value.comment:
            res["knora-api:valueHasComment"] = value.comment
        if value.permissions:
            if perm := self.permissions_lookup.get(value.permissions):
                res["knora-api:hasPermissions"] = str(perm)
            else:
                raise PermissionNotExistsError(f"Could not find permissions for value: {value.permissions}")
        return res


def _add_optional_permission_triple(
    value: XMLValue | IIIFUriInfo, val_bn: BNode, g: Graph, permissions_lookup: dict[str, Permissions]
) -> None:
    if value.permissions:
        if not (per := permissions_lookup.get(value.permissions)):
            raise PermissionNotExistsError(f"Could not find permissions for value: {value.permissions}")
        g.add((val_bn, KNORA_API.hasPermissions, Literal(str(per))))


def _make_iiif_uri_value(iiif_uri: IIIFUriInfo, res_bnode: BNode, permissions_lookup: dict[str, Permissions]) -> Graph:
    g = Graph()
    iiif_bn = BNode()
    g.add((res_bnode, KNORA_API.hasStillImageFileValue, iiif_bn))
    g.add((iiif_bn, RDF.type, KNORA_API.StillImageExternalFileValue))
    g.add((iiif_bn, KNORA_API.fileValueHasExternalUrl, Literal(iiif_uri.value)))
    _add_optional_permission_triple(iiif_uri, iiif_bn, g, permissions_lookup)
    return g


def _make_bitstream_file_value(bitstream_info: BitstreamInfo) -> dict[str, Any]:
    local_file = Path(bitstream_info.local_file)
    file_ending = local_file.suffix[1:]
    match file_ending.lower():
        case "zip" | "tar" | "gz" | "z" | "tgz" | "gzip" | "7z":
            prop = "knora-api:hasArchiveFileValue"
            value_type = "ArchiveFileValue"
        case "mp3" | "wav":
            prop = "knora-api:hasAudioFileValue"
            value_type = "AudioFileValue"
        case "pdf" | "doc" | "docx" | "xls" | "xlsx" | "ppt" | "pptx":
            prop = "knora-api:hasDocumentFileValue"
            value_type = "DocumentFileValue"
        case "mp4":
            prop = "knora-api:hasMovingImageFileValue"
            value_type = "MovingImageFileValue"
        # jpx is the extension of the files returned by dsp-ingest
        case "jpg" | "jpeg" | "jp2" | "png" | "tif" | "tiff" | "jpx":
            prop = "knora-api:hasStillImageFileValue"
            value_type = "StillImageFileValue"
        case "odd" | "rng" | "txt" | "xml" | "xsd" | "xsl" | "csv":
            prop = "knora-api:hasTextFileValue"
            value_type = "TextFileValue"
        case _:
            raise BaseError(f"Unknown file ending '{file_ending}' for file '{local_file}'")
    file_value = {
        "@type": f"knora-api:{value_type}",
        "knora-api:fileValueHasFilename": bitstream_info.internal_file_name,
    }
    if bitstream_info.permissions:
        file_value["knora-api:hasPermissions"] = str(bitstream_info.permissions)
    return {prop: file_value}


def _to_boolean(s: str | int | bool) -> bool:
    match s:
        case "True" | "true" | "1" | 1 | True:
            return True
        case "False" | "false" | "0" | 0 | False:
            return False
        case _:
            raise BaseError(f"Could not parse boolean value: {s}")


def _make_color_value(value: XMLValue) -> dict[str, Any]:
    return {
        "@type": "knora-api:ColorValue",
        "knora-api:colorValueAsColor": value.value,
    }


def _make_date_value(value: XMLValue) -> dict[str, Any]:
    string_value = _assert_is_string(value.value)
    date = parse_date_string(string_value)
    res: dict[str, Any] = {
        "@type": "knora-api:DateValue",
        "knora-api:dateValueHasStartYear": date.start.year,
    }
    if month := date.start.month:
        res["knora-api:dateValueHasStartMonth"] = month
    if day := date.start.day:
        res["knora-api:dateValueHasStartDay"] = day
    if era := date.start.era:
        res["knora-api:dateValueHasStartEra"] = era.value
    if calendar := date.calendar:
        res["knora-api:dateValueHasCalendar"] = calendar.value
    if date.end:
        res["knora-api:dateValueHasEndYear"] = date.end.year
        if month := date.end.month:
            res["knora-api:dateValueHasEndMonth"] = month
        if day := date.end.day:
            res["knora-api:dateValueHasEndDay"] = day
        if era := date.end.era:
            res["knora-api:dateValueHasEndEra"] = era.value
    return res


def _make_decimal_value(value: XMLValue) -> dict[str, Any]:
    s = _assert_is_string(value.value)
    return {
        "@type": "knora-api:DecimalValue",
        "knora-api:decimalValueAsDecimal": {
            "@type": "xsd:decimal",
            "@value": str(float(s)),
        },
    }


def _make_geometry_value(value: XMLValue) -> dict[str, Any]:
    s = _assert_is_string(value.value)
    # this removes all whitespaces from the embedded json string
    encoded_value = json.dumps(json.loads(s))
    return {
        "@type": "knora-api:GeomValue",
        "knora-api:geometryValueAsGeometry": encoded_value,
    }


def _make_geoname_value(value: XMLValue) -> dict[str, Any]:
    return {
        "@type": "knora-api:GeonameValue",
        "knora-api:geonameValueAsGeonameCode": value.value,
    }


def _make_boolean_prop(
    prop: XMLProperty, res_bn: BNode, prop_name: URIRef, permissions_lookup: dict[str, Permissions]
) -> Graph:
    g = Graph()
    for value in prop.values:
        single_val_bn = BNode()
        g.add((res_bn, prop_name, single_val_bn))
        g += _make_boolean_value(value, single_val_bn, permissions_lookup)
    return g


def _make_boolean_value(value: XMLValue, val_bn: BNode, permissions_lookup: dict[str, Permissions]) -> Graph:
    s = _assert_is_string(value.value)
    g = Graph()
    g.add((val_bn, RDF.type, KNORA_API.BooleanValue))
    g.add((val_bn, KNORA_API.booleanValueAsBoolean, Literal(_to_boolean(s))))
    _add_optional_permission_triple(value, val_bn, g, permissions_lookup)
    if value.comment:
        g.add((val_bn, KNORA_API.valueHasComment, Literal(value.comment)))
    return g


def _make_integer_prop(
    prop: XMLProperty, res_bn: BNode, prop_name: URIRef, permissions_lookup: dict[str, Permissions]
) -> Graph:
    g = Graph()
    for value in prop.values:
        single_val_bn = BNode()
        g.add((res_bn, prop_name, single_val_bn))
        g += _make_integer_value(value, single_val_bn, permissions_lookup)
    return g


def _make_integer_value(value: XMLValue, val_bn: BNode, permissions_lookup: dict[str, Permissions]) -> Graph:
    s = _assert_is_string(value.value)
    g = Graph()
    g.add((val_bn, RDF.type, KNORA_API.IntValue))
    g.add((val_bn, KNORA_API.intValueAsInt, Literal(int(s))))
    _add_optional_permission_triple(value, val_bn, g, permissions_lookup)
    if value.comment:
        g.add((val_bn, KNORA_API.valueHasComment, Literal(value.comment)))
    return g


def _make_interval_value(value: XMLValue) -> dict[str, Any]:
    s = _assert_is_string(value.value)
    match s.split(":", 1):
        case [start, end]:
            return {
                "@type": "knora-api:IntervalValue",
                "knora-api:intervalValueHasStart": {
                    "@type": "xsd:decimal",
                    "@value": str(float(start)),
                },
                "knora-api:intervalValueHasEnd": {
                    "@type": "xsd:decimal",
                    "@value": str(float(end)),
                },
            }
        case _:
            raise BaseError(f"Could not parse interval value: {s}")


def _make_link_value(value: XMLValue, iri_resolver: IriResolver) -> dict[str, Any]:
    s = _assert_is_string(value.value)
    if is_resource_iri(s):
        iri = s
    elif resolved_iri := iri_resolver.get(s):
        iri = resolved_iri
    else:
        msg = (
            f"Could not find the ID {s} in the id2iri mapping. "
            f"This is probably because the resource '{s}' could not be created. "
            f"See {WARNINGS_SAVEPATH} for more information."
        )
        raise BaseError(msg)
    return {
        "@type": "knora-api:LinkValue",
        "knora-api:linkValueHasTargetIri": {
            "@id": iri,
        },
    }


def _make_list_value(value: XMLValue, iri_lookup: dict[str, str]) -> dict[str, Any]:
    s = _assert_is_string(value.value)
    if iri := iri_lookup.get(s):
        return {
            "@type": "knora-api:ListValue",
            "knora-api:listValueAsListNode": {
                "@id": iri,
            },
        }
    else:
        msg = (
            f"Could not resolve list node ID '{s}' to IRI. "
            f"This is probably because the list node '{s}' does not exist on the server."
        )
        raise BaseError(msg)


def _make_text_value(value: XMLValue, iri_resolver: IriResolver) -> dict[str, Any]:
    match value.value:
        case str() as s:
            return {
                "@type": "knora-api:TextValue",
                "knora-api:valueAsString": s,
            }
        case FormattedTextValue() as xml:
            xml_with_iris = xml.with_iris(iri_resolver)
            return {
                "@type": "knora-api:TextValue",
                "knora-api:textValueAsXml": xml_with_iris.as_xml(),
                "knora-api:textValueHasMapping": {
                    "@id": "http://rdfh.ch/standoff/mappings/StandardMapping",
                },
            }
        case _:
            assert_never(value.value)


def _make_time_value(value: XMLValue) -> dict[str, Any]:
    return {
        "@type": "knora-api:TimeValue",
        "knora-api:timeValueAsTimeStamp": {
            "@type": "xsd:dateTimeStamp",
            "@value": value.value,
        },
    }


def _serialise_uri_prop(prop: XMLProperty, permissions_lookup: dict[str, Permissions]) -> dict[str, Any]:
    uri_values: list[SerialiseValue] = []
    for v in prop.values:
        if v.permissions:
            if not (per := permissions_lookup.get(v.permissions)):
                raise PermissionNotExistsError(f"Could not find permissions for value: {v.permissions}")
            permission = str(per)
        else:
            permission = None
        uri = cast(str, v.value)
        uri_values.append(
            SerialiseURI(
                value=uri,
                permissions=permission,
                comment=v.comment,
            )
        )
    prop_serialise = SerialiseProperty(
        property_name=prop.name,
        values=uri_values,
    )
    return prop_serialise.serialise()


def _assert_is_string(value: str | FormattedTextValue) -> str:
    match value:
        case str() as s:
            return s
        case FormattedTextValue() as xml:
            raise BaseError(f"Expected string value, but got XML value: {xml.as_xml()}")
        case _:
            assert_never(value)
